<?php
/**
 * @version SOFORT Gateway 3.3.0 - $Date: 2012-07-09 14:07:58 +0200 (Mo, 09. Jul 2012) $
 * @author SOFORT AG (integration@sofort.com)
 * @link http://www.sofort.com/
 *
 * Copyright (c) 2012 SOFORT AG
 *
 * Released under the GNU General Public License (Version 2)
 * [http://www.gnu.org/licenses/gpl-2.0.html]
 *
 * $Id: sofortOrderSynchronisation.php 4663 2012-07-09 12:07:58Z dehn $
 */

require_once(DIR_FS_CATALOG."ext/modules/payment/sofort/ressources/scripts/sofortOrderShopTools.php");
require_once(DIR_FS_CATALOG."ext/modules/payment/sofort/helperFunctions.php");

class sofortOrderSynchronisation {
	
	/**
	 * if cart is edited at shop, send new data to SOFORT
	 * @param string $transactionId
	 * @param array	 $articles
	 * @param string $comment
	 */
	public function editArticlesSofort($transactionId, $articles, $comment = '') {
		$PnagInvoice = new PnagInvoice(shopGetConfigKey(), $transactionId);
		$sofortArticles = array();
		$orderId = '';
		$taxLow = 0;
		$taxHigh = 0;
		$subtotal = 0;
		$shipping = 0;
		$discount = array();
		$agio = array();
		
		foreach($articles as $article) {
			if ($orderId == '') $orderId = $article['articleOrdersId'];
			
			if($article['articleQuantity'] != 'delete'){
				array_push($sofortArticles, array(
					'itemId'		=> $article['articleId'],
					'productNumber' => $article['articleNumber'],
					'title'			=> HelperFunctions::convertEncoding($article['articleTitle'],3),
					'description'	=> HelperFunctions::convertEncoding($article['articleDescription'],3),
					'quantity'		=> $article['articleQuantity'],
					'unitPrice'		=> number_format($article['articlePrice'], 2, '.', ''),
					'tax'			=> number_format($article['articleTax'], 2, '.', ''))
				);
			}
			
			switch($article['articleType']){
				case 'shipping': 	$shipping = $article['articlePrice'];
									break;
				case 'discount':	$splitItemId = explode('|', $article['articleId']);
									$discountClass = $splitItemId[1];
									array_push($discount, array(
											'class' => $discountClass,
											'value' => $article['articlePrice']));
									break;
				case 'agio':		$splitItemId = explode('|', $article['articleId']);
									$agioClass = $splitItemId[1];
									array_push($agio , array(
											'class' => $agioClass,
											'value' => $article['articlePrice']));
									break;
				case 'product':		$subtotal += ($article['articleQuantity'] * $article['articlePrice']);
									break;
			}
			
			switch ($article['articleTax']){
				case 7:		$taxLow  += ($article['articleQuantity']*$article['articlePrice']);
							break;
				case 19:	$taxHigh += ($article['articleQuantity']*$article['articlePrice']);
							break;
			}
		}
		
		$lastShopTotal = HelperFunctions::getLastFieldValueFromSofortTable($orderId, 'amount', true);
		$time = date("d.m.Y, G:i:s");
		$errors = $PnagInvoice->updateInvoice($transactionId, $sofortArticles, $comment);
		$warnings = $PnagInvoice->getWarnings();
		
		if($errors){
			return array(
					'errors'   => $errors,
					'warnings' => $warnings
			);
		}
		
		$PnagInvoice->refreshTransactionData();
		$orderStatus = shopDbQuery('SELECT orders_status FROM '.TABLE_ORDERS.' WHERE orders_id = "'.(int)$orderId.'"');
		$orderStatus = shopDbFetchArray($orderStatus);
		$invoiceArticles = $PnagInvoice->getItems();
		
		if(is_array($invoiceArticles)){
			if (HelperFunctions::checkIfNewTotal($PnagInvoice->getAmount(), $orderId)){
				$this->_insertNewTotalCommentToHistory($orderId, $orderStatus['orders_status'], $time, $PnagInvoice, $lastShopTotal);
			}
			
			$this->editArticlesShop($PnagInvoice, $orderId);
		} else {
			return array(
					'errors'   => 'Undefined Error (probably unable to connect to the SOFORT-Server)'
			);
		}
		
		return false;
	}
	
	
	/**
	 * cart was edited at SOFORT-backend, apply changes in shop
	 * @param PnagInvoice $PnagInvoice
	 */
	public function editArticlesShop(PnagInvoice $PnagInvoice, $orderNumber) {
		$lng = $PnagInvoice->getLanguageCode();
		$newAmount = $PnagInvoice->getAmount();
		$invoiceArticles = $PnagInvoice->getItems();
		$sofortIdArray = array();
		
		foreach ($invoiceArticles as $article) {
			$getTotalItems = explode('|', $article->itemId);
			
			//discount/agio/shipping/total-modules
			if(count($getTotalItems) > '1'){
				if($getTotalItems[0]=='discount' || $getTotalItems[0]=='agio'){
					$sofortIdArray[$article->itemId] = $article->itemId;
					$sofortArticleArray[$article->itemId] = $article;
				} else {
					$sofortIdArray[$getTotalItems[0]] = $getTotalItems[0];
					$sofortArticleArray[$getTotalItems[0]] = $article;
				}
			//normal product
			} else {
				$ordersProductsId = $this->_getOrderProductsId($article->itemId, $orderNumber);
				
				if (!$ordersProductsId) {
					$PnagInvoice->logError("Critical Error! At least one OrdersProductsId for orderID '$orderNumber' was not found in shop. Article with
						item-ID '".$article->itemId."' does not exist in shop. Syncronisation will be aborted. Compare this order and prices with the order
						at SOFORT for differences. Check also comments in order-history and status of this order in Shop. Data at SOFORT are normally
						not affected and correct. Affected invoice-article: ".print_r($article, true));
					return false; //prevent sync
				}
				
				$sofortIdArray[$ordersProductsId] = $ordersProductsId;
				$sofortArticleArray[$ordersProductsId] = $article;
			}
		}
		
		$shopProductsQuery = shopDbQuery("SELECT orders_products_id FROM ".TABLE_ORDERS_PRODUCTS." WHERE orders_id = '".(int)$orderNumber."'");
		
		while ($shopProductsResult = shopDbFetchArray($shopProductsQuery)) $shopArticleArray[] = $shopProductsResult['orders_products_id'];
		
		$taxLow = 0;
		$taxHigh = 0;
		$subtotal = 0;
		
		foreach ($shopArticleArray as $shopArticle){
			if (!in_array($shopArticle,$sofortIdArray)){
				$this->_sofortRestock($this->_getItemId($shopArticle, $orderNumber), $orderNumber, 0);
				$this->_deleteShopOrderArticle($shopArticle, $PnagInvoice->getStatusReason());
			} else {
				$qty = $sofortArticleArray[$shopArticle]->quantity;
				$price = $sofortArticleArray[$shopArticle]->unitPrice;
				$itemId = $sofortArticleArray[$shopArticle]->itemId;
				$this->_sofortRestock($itemId, $orderNumber, $qty);
				$this->_updateShopOrderArticle($shopArticle, $qty, $price, $PnagInvoice->getStatusReason(), $itemId, $orderNumber);
				
				if ($sofortArticleArray[$shopArticle]->tax == '7.00'){
					$taxLow  += ($sofortArticleArray[$shopArticle]->quantity * $sofortArticleArray[$shopArticle]->unitPrice);
				} elseif ($sofortArticleArray[$shopArticle]->tax == '19.00'){
					$taxHigh += ($sofortArticleArray[$shopArticle]->quantity * $sofortArticleArray[$shopArticle]->unitPrice);
				}
				
				$subtotal += ($sofortArticleArray[$shopArticle]->quantity * $sofortArticleArray[$shopArticle]->unitPrice);
			}
		}
		
		$shipping = 0;
		$discount = array();
		$agio = array();
		$sofortTotals = array('ot_total', 'ot_subtotal', 'ot_shipping');
		
		foreach ($sofortIdArray as $sofortId){
			if (!in_array($sofortId, $shopArticleArray)){
				$splitItemId = explode('|', $sofortId);
				
				switch($splitItemId[0]){
					case 'shipping': 	$shipping = $sofortArticleArray[$sofortId]->unitPrice;
										break;
					case 'discount':	$discountClass = $splitItemId[1];
										array_push($discount, array(
												'class' => $discountClass,
												'value' => $sofortArticleArray[$sofortId]->unitPrice
												)
										);
										array_push($sofortTotals, $discountClass);
										break;
					case 'agio':		$agioClass = $splitItemId[1];
										array_push($agio , array(
												'class' => $agioClass,
												'value' => $sofortArticleArray[$sofortId]->unitPrice
												)
										);
										array_push($sofortTotals, $agioClass);
										break;
					default:			$this->_sofortRestock($sofortArticleArray[$sofortId]->itemId, $orderNumber, $sofortArticleArray[$sofortId]->quantity);
										$this->_insertShopOrderArticle($sofortArticleArray[$sofortId], $orderNumber, $lng);
										$subtotal += ($sofortArticleArray[$sofortId]->quantity * $sofortArticleArray[$sofortId]->unitPrice);
										break;
				}
				
				if ($sofortArticleArray[$sofortId]->tax == '7.00'){
					$taxLow += ($sofortArticleArray[$sofortId]->unitPrice*$sofortArticleArray[$sofortId]->quantity);
				} elseif ($sofortArticleArray[$sofortId]->tax == '19.00'){
					$taxHigh += ($sofortArticleArray[$sofortId]->unitPrice*$sofortArticleArray[$sofortId]->quantity);
				}
			}
		}
		
		$totalsQuery = shopDbQuery('SELECT class, value, title FROM '.TABLE_ORDERS_TOTAL.' WHERE orders_id = "'.shopDbInput($orderNumber).'"');
		
		while ($totalsResult = shopDbFetchArray($totalsQuery)){
			if (!in_array($totalsResult['class'], $sofortTotals)) {
				$this->_checkAndUpdateTotal($orderNumber, $totalsResult['class'], '', '', '0', '', '');
			}
		}
		
		$status = $PnagInvoice->getStatusReason();
		$this->_updateShopTotals($taxLow, $taxHigh, $subtotal, $newAmount, $orderNumber, $shipping, $discount, $agio, $status);
	}
	
	
	/**
	 * get SOFORT-ItemID from table sofort_products
	 * @param int $ordersProductsId
	 * @param int $ordersId
	 */
	protected function _getItemId ($ordersProductsId, $ordersId) {
		$query = "SELECT item_id FROM sofort_products WHERE orders_products_id = ".shopDbInput($ordersProductsId);
		$product = shopDbCheckAndFetchOne($query);
		
		return $product['item_id'];
	}
	
	
	/**
	 * get latest item-quantity
	 * @param int $itemId
	 * @param int $ordersId
	 */
	protected function _getLatestQuantity($itemId, $ordersId){
		$qry = "SELECT products_quantity FROM ".TABLE_ORDERS_PRODUCTS." WHERE orders_products_id = '".shopDbInput($this->_getOrderProductsId($itemId, $ordersId))."'";
		$res = shopDbCheckAndFetchOne($qry);
		
		return $res['products_quantity'];
	}
	
	
	/**
	 * get number of different products
	 * @param int $ordersId
	 */
	protected function _getNumberOfOrderProducts ($ordersId) {
		$query = "SELECT order_products_id FROM ".TABLE_ORDERS_PRODUCTS." WHERE orders_id = ".(int)$ordersId;
		$number = shopDbGetNumRows($query);
		
		return $number;
	}
	
	
	/**
	 * get orders_products_id from table sofort_products
	 * @param int $itemId
	 * @param int $ordersId
	 */
	protected function _getOrderProductsId ($itemId, $ordersId) {
		$query = shopDbQuery("SELECT orders_products_id FROM sofort_products WHERE item_id = '".shopDbInput($itemId)."' AND orders_id = '".(int)$ordersId."'");
		$product = shopDbFetchArray($query);
		
		return $product['orders_products_id'];
	}
	
	
	/**
	 * update quantity and price in shop order table
	 * @param int	 $ordersProductsId
	 * @param mixed	 $quantity
	 * @param float	 $unitPrice
	 * @param string $status
	 */
	protected function _updateShopOrderArticle($ordersProductsId, $quantity, $unitPrice, $status, $itemId, $orderId){
		if ($quantity == 'delete'){
			// article was marked for removal which was already handled in editArticlesSofort()
		} elseif ($quantity == 0){
			$this->_deleteShopOrderArticle($ordersProductsId, $status);
		} else {
			$tax = shopDbFetchArray(shopDbQuery('SELECT products_tax FROM '.TABLE_ORDERS_PRODUCTS.' WHERE orders_products_id = "'.shopDbInput($ordersProductsId).'"'));
			$finalPrice = $this->_deleteTax($unitPrice, $tax['products_tax']);
			$productPrice = $this->_getProductPrice($finalPrice, $ordersProductsId, $orderId, $itemId);
			shopDbQuery("UPDATE ".TABLE_ORDERS_PRODUCTS." SET products_quantity = '".(int)$quantity."', products_price = '".shopDbInput($productPrice)."', final_price = '".shopDbInput($finalPrice)."' WHERE orders_products_id ='".shopDbInput($ordersProductsId)."'");
		}
	}
	
	
	/**
	 * update shop totals
	 * @param float	 $taxLow
	 * @param float	 $taxHigh
	 * @param float	 $subtotal
	 * @param float	 $newAmount
	 * @param int	 $ordersId
	 * @param array	 $shipping
	 * @param array	 $discount
	 * @param array	 $agio
	 * @param string $status
	 * @param string $currency
	 */
	protected function _updateShopTotals($taxLow, $taxHigh, $subtotal, $newAmount, $ordersId, $shipping, $discount, $agio, $status, $currency = 'EUR'){
		if ($status == 'canceled' || $status == 'confirmation_period_expired' || $status == 'refunded') return; // in case of cancellation keep old data for replicability
		
		shopDbQuery('UPDATE '.TABLE_ORDERS_TOTAL.' SET value = "'.shopDbInput($newAmount).'", text = "<b>'.shopDbInput(number_format($newAmount,2,",",".")).' '.shopDbInput($currency).'</b>" WHERE orders_id = "'.(int)$ordersId.'" AND class = "ot_total"');
		shopDbQuery('UPDATE '.TABLE_ORDERS_TOTAL.' SET value = "'.shopDbInput($subtotal).'" , text = "'.shopDbInput(number_format($subtotal,2,",",".")).' '.shopDbInput($currency).'" WHERE orders_id = "'.(int)$ordersId.'" AND class = "ot_subtotal"');
		
		$shippingText = number_format($shipping ,2,",",".").' '.$currency;
		$this->_checkAndUpdateTotal($ordersId, 'ot_shipping', 'Versandkosten:', $shippingText, $shipping, MODULE_ORDER_TOTAL_SHIPPING_SORT_ORDER);
		$taxLowValue = 0;
		
		if ($taxLow  != 0) $taxLowValue  = (($taxLow /107)* 7);
		
		$taxLowText = number_format($taxLowValue ,2,",",".").' '.$currency;
		$this->_checkAndUpdateTotal($ordersId, 'ot_tax', 'zzgl. 7% MwSt.:', $taxLowText, $taxLowValue, MODULE_ORDER_TOTAL_TAX_SORT_ORDER, ' AND title LIKE  "%7%"');
		$taxHighValue = 0;
		
		if ($taxHigh != 0) $taxHighValue = (($taxHigh/119)*19);
		
		$taxHighText = number_format($taxHighValue ,2,",",".").' '.$currency;
		$this->_checkAndUpdateTotal($ordersId, 'ot_tax', 'zzgl. 19% MwSt.:', $taxHighText, $taxHighValue, MODULE_ORDER_TOTAL_TAX_SORT_ORDER, ' AND title LIKE  "%19%"');
		$nettoValue = $newAmount - $taxLowValue - $taxHighValue;
		$nettoText = number_format($nettoValue ,2,",",".").' '.$currency;
		$this->_checkAndUpdateNetto($ordersId, $nettoValue, $newAmount, $nettoText);
		
		if(count($discount)!=0){
			foreach ($discount as $position){
				$text = number_format($position['value'],2,',','.')." ".$currency;
				$modulName = substr($position['class'], 3); //delete "ot_"
				$this->_checkAndUpdateTotal($ordersId, $position['class'], 'Rabatt:', $text, $position['value'], constant('MODULE_ORDER_TOTAL_'.strtoupper($modulName).'_SORT_ORDER'));
			}
		}
		
		if(count($agio)!=0){
			foreach ($agio as $position){
				$text = number_format($position['value'],2,",",".").' '.$currency;
				$modulName = substr($position['class'], 3); //delete "ot_"
				$this->_checkAndUpdateTotal($ordersId, $position['class'], 'Zuschlag:', $text, $position['value'], constant('MODULE_ORDER_TOTAL_'.strtoupper($modulName).'_SORT_ORDER'));
			}
		}
	}
	
	
	/**
	 * Checks if total exists; updates or deletes if true; inserts if false
	 * @param int	 $ordersId
	 * @param string $class
	 * @param string $title
	 * @param string $text
	 * @param float	 $value
	 * @param string $addCond - has to be already escaped!
	 */
	protected function _checkAndUpdateTotal($ordersId, $class, $title, $text, $value, $sortOrder, $addCond = ''){
		if(shopDbNumRows(shopDbQuery('SELECT * FROM '.TABLE_ORDERS_TOTAL.' WHERE orders_id = "'.(int)$ordersId.'" AND class = "'.shopDbInput($class).'"'.$addCond)) != 0) {
			if ($value != 0 || $class == 'ot_shipping'){
				shopDbQuery('UPDATE '.TABLE_ORDERS_TOTAL.' SET value = "'.shopDbInput($value).'" , text = "'.shopDbInput($text).'" WHERE orders_id = "'.(int)$ordersId.'" AND class = "'.shopDbInput($class).'"'.$addCond);
			} else {
				shopDbQuery('DELETE FROM '.TABLE_ORDERS_TOTAL.' WHERE orders_id = "'.(int)$ordersId.'" AND class = "'.shopDbInput($class).'"'.$addCond);
			}
		} else {
			if ($value != 0 || $class == 'ot_shipping'){
				$data = array(
						'value'		=> $value,
						'text'		=> $text,
						'orders_id' => $ordersId,
						'class'		=> $class,
						'title'		=> $title,
						'sort_order'=> $sortOrder
				);
				shopDbPerform(TABLE_ORDERS_TOTAL, $data);
			}
		}
	}
	
	
	/**
	 * Checks if netto-string exists; updates if true; inserts if false and bruttoAmount >= 150
	 * @param int	 $ordersId
	 * @param string $text
	 * @param float $nettoAmount
	 * @param float $bruttoAmount
	 */
	protected function _checkAndUpdateNetto($ordersId, $nettoAmount, $bruttoAmount, $text){
		if(shopDbNumRows(shopDbQuery('SELECT * FROM '.TABLE_ORDERS_TOTAL.' WHERE orders_id = "'.$ordersId.'" AND class LIKE "%netto%"')) != 0) {
			shopDbQuery('UPDATE '.TABLE_ORDERS_TOTAL.' SET value = "'.$nettoAmount.'" , text = "'.$text.'" WHERE orders_id = "'.$ordersId.'" AND class LIKE "%netto%"');
		} else if ($bruttoAmount >= 150) {
			$data = array(
					'value'		=> $nettoAmount,
					'text'		=> $text,
					'orders_id' => $ordersId,
					'class'		=> 'ot_total_netto',
					'title'		=> 'Summe (netto):',
					'sort_order'=> (MODULE_ORDER_TOTAL_TAX_SORT_ORDER - 1)
			);
			shopDbPerform(TABLE_ORDERS_TOTAL, $data);
		}
	}
	
	
	/**
	 * delete article from shop order
	 * @param int	 $ordersProductsId
	 * @param string $status
	 */
	protected function _deleteShopOrderArticle($ordersProductsId, $status){
		if ($status == 'canceled' || $status == 'refunded' || $status == 'confirmation_period_expired'){
			// in case of cancellation keep old data for replicability
		} else {
			$query = "DELETE FROM ".TABLE_ORDERS_PRODUCTS." WHERE orders_products_id = '".shopDbInput($ordersProductsId)."'";
			shopDbQuery($query);
			$query = "DELETE FROM ".TABLE_ORDERS_PRODUCTS_ATTRIBUTES." WHERE orders_products_id = '".shopDbInput($ordersProductsId)."'";
			shopDbQuery($query);
		}
	}
	
	
	/**
	 * insert article in shop order (e.g. during an undo operation)
	 * @param object $sofortItem
	 * @param int	 $ordersId
	 * @param string $lng
	 */
	protected function _insertShopOrderArticle($sofortItem, $ordersId, $lng){
		$itemId = $sofortItem->itemId;
		$splitItemId = explode('{',$itemId);
		$productId = $splitItemId[0];
		
		if(count($splitItemId) == '1'){
			$hasAttributes = false;
		} else {
			$hasAttributes = true;
			
			for ($i=1;$i<count($splitItemId);++$i){
				$attrId = explode('}',$splitItemId[$i]);
				$attributes[] = array(
						'optionsId'		  => $attrId[0],
						'optionsValuesId' => $attrId[1]
				);
			}
		}
		
		$data = array(
				'orders_id'			=> $ordersId,
				'products_id'		=> $productId,
				'products_model'	=> $sofortItem->productNumber,
				'products_name'		=> HelperFunctions::convertEncoding($sofortItem->title,2),
				'products_price'	=> $this->_deleteTax($sofortItem->unitPrice, $sofortItem->tax),
				'final_price'		=> $this->_deleteTax($sofortItem->unitPrice, $sofortItem->tax),
				'products_tax'		=> $sofortItem->tax,
				'products_quantity' => $sofortItem->quantity,
		);
		shopDbPerform(TABLE_ORDERS_PRODUCTS, $data);
		$insertId = tep_db_insert_id();
		shopDbQuery('UPDATE sofort_products SET orders_products_id ="'.(int)$insertId.'" WHERE orders_id = "'.(int)$ordersId.'" AND item_id = "'.shopDbInput($itemId).'"');
		
		if($hasAttributes) {
			$lngId = shopDbFetchArray(shopDbQuery("SELECT languages_id FROM ".TABLE_LANGUAGES." WHERE code = '".shopDbInput($lng)."'"));
			$attributeCostExists = false;
			
			foreach($attributes as $attribute){
				$queryTpa = shopDbQuery("SELECT options_values_price, price_prefix FROM ".TABLE_PRODUCTS_ATTRIBUTES." WHERE products_id ='".shopDbInput($productId)."' AND options_id = '".shopDbInput($attribute['optionsId'])."' AND options_values_id ='".shopDbInput($attribute['optionsValuesId'])."'");
				$resultTpa = shopDbFetchArray($queryTpa);
				$queryTpo = shopDbQuery("SELECT products_options_name FROM ".TABLE_PRODUCTS_OPTIONS." WHERE products_options_id = '".shopDbInput($attribute['optionsId'])."' AND language_id = '".shopDbInput($lngId['languages_id'])."'");
				$resultTpo = shopDbFetchArray($queryTpo);
				$queryTpov = shopDbQuery("SELECT products_options_values_name FROM ".TABLE_PRODUCTS_OPTIONS_VALUES." WHERE products_options_values_id = '".shopDbInput($attribute['optionsValuesId'])."' AND language_id = '".shopDbInput($lngId['languages_id'])."'");
				$resultTpov = shopDbFetchArray($queryTpov);
				$data = array(
						'orders_id'				  => $ordersId,
						'orders_products_id'	  => $insertId,
						'products_options'		  => $resultTpo['products_options_name'],
						'products_options_values' => $resultTpov['products_options_values_name'],
						'options_values_price'	  => $resultTpa['options_values_price'],
						'price_prefix'			  => $resultTpa['price_prefix']
				);
				shopDbPerform(TABLE_ORDERS_PRODUCTS_ATTRIBUTES, $data);
				
				if ($resultTpa['options_values_price'] != 0) $attributeCostExists = true;
			}
			
			if ($attributeCostExists) {
				$productPrice = $this->_getProductPrice($this->_deleteTax($sofortItem->unitPrice, $sofortItem->tax), $ordersProductsId, $orderId, $itemId);
				shopDbQuery('UPDATE '.TABLE_ORDERS_PRODUCTS.' SET product_price = "'.shopDbInput($productPrice).'" WHERE orders_id = "'.(int)$ordersId.'" AND orders_products_id = "'.(int)$insertId.'"');
			}
		}
	}
	
	
	/**
	 * restock shop articles during cartsynchronization, cancelation etc.
	 * @param string $itemId
	 * @param int	 $ordersId
	 * @param int	 $newQty
	 * @param int	 $oldQty
	 */
	protected function _sofortRestock($itemId, $ordersId, $newQty, $oldQty = 'aaaa') {
		if(!is_numeric($oldQty)) $oldQty = $this->_getLatestQuantity($itemId, $ordersId);
		
		$splitItemId = explode('{',$itemId);
		$productId = $splitItemId[0];
		$diff = $oldQty - $newQty;
		$updateTP = "UPDATE ".TABLE_PRODUCTS." SET products_quantity = products_quantity + ".shopDbInput($diff)." WHERE products_id = '".shopDbInput($productId)."'";
		shopDbQuery($updateTP);
	}
	
	
	/**
	 * inserts a "new total" comment into shop order status history
	 * @param int		  $orderId
	 * @param string	  $status
	 * @param date		  $time (or false, if it should not be set)
	 * @param PnagInvoice $PnagInvoice
	 * @param float		  $lastShopTotal
	 */
	protected function _insertNewTotalCommentToHistory($orderId, $status, $time, PnagInvoice $PnagInvoice, $lastShopTotal){
		$newTotal = $PnagInvoice->getAmount();
		
		$comment = ($newTotal > $lastShopTotal) ? SOFORT_SR_TRANSLATE_CART_RESET : SOFORT_SR_TRANSLATE_CART_EDITED;
		
		$comment .= ' '.SOFORT_SR_TRANSLATE_CURRENT_TOTAL.' '.$newTotal.' Euro';
		
		if ($time) $comment .= ' '.SOFORT_TRANSLATE_TIME.': '.$time;
		
		$sqlDataArray = array(
				'orders_id'			=> (int)$orderId,
				'orders_status_id'	=> $status,
				'date_added'		=> 'now()',
				'customer_notified' => 0,
				'comments'			=> $comment,
		);
		shopDbPerform(TABLE_ORDERS_STATUS_HISTORY, $sqlDataArray);
		
		//save all relevant data in table sofort_notification
		//because we dont get a notification in this situation
		$dbEntries = array(
			'amount' => $newTotal,
			'status' => $PnagInvoice->getStatus(),
			'status_reason' => $PnagInvoice->getStatusReason(),
			'status_id' => $PnagInvoice->getState(),
		);
		HelperFunctions::insertSofortOrdersNotification($PnagInvoice->getTransactionId(), $dbEntries);
	}
	
	
	/**
	 * remove amount of given tax from given price (e.g. price 119, tax 19 -> returns 100)
	 * @param float $price
	 * @param float $tax
	 */
	protected function _deleteTax ($price, $tax) {
		return ($price / (1 + ($tax/100)));
	}
	
	
	/**
	 * get productPrice for shop
	 * finalPrice = productPrice + attributesPrice - SOFORT only needs the finalPrice, but the Shop differentiates
	 * @param float	 $finalPrice
	 * @param int	 $ordersProductsId
	 * @param int	 $orderId
	 * @param string $itemId
	 * @return float $productPrice
	 */
	protected function _getProductPrice($finalPrice, $ordersProductsId, $orderId, $itemId) {
		$splitItemId = explode('{',$itemId);
		$hasAttributes = (count($splitItemId) == '1') ? false : true;
		$attributesNetAmount = 0;
		$checkAmount = 0;
		
		if ($hasAttributes) {
			for ($i=1; $i<count($splitItemId); ++$i) {
				$attributeIds = explode('}',$splitItemId[$i]);
				$checkQuery = shopDbQuery("SELECT options_values_price, price_prefix FROM ".TABLE_PRODUCTS_ATTRIBUTES." WHERE products_id ='".$splitItemId[0]."' AND options_id = '".$attributeIds[0]."' AND options_values_id = '".$attributeIds[1]."'");
				$checkResult = shopDbFetchArray($checkQuery);
				
				switch ($checkRes['price_prefix']) {
					case '+':	$checkAmount += $checkResult['options_values_price']; break;
					case '-':	$checkAmount -= $checkResult['options_values_price']; break;
					default:	break;
				}
			}
			
			$prodQry = shopDbQuery("SELECT products_price FROM ".TABLE_PRODUCTS." WHERE products_id = '".$splitItemId[0]."'");
			$prodRes = shopDbFetchArray($prodQry);
			$checkAmount += $prodRes['products_price'];
			$attributesPrices = shopDbQuery("SELECT options_values_price, price_prefix FROM ".TABLE_ORDERS_PRODUCTS_ATTRIBUTES." WHERE orders_products_id = '".shopDbInput($ordersProductsId)."' AND orders_id = '".shopDbInput($orderId)."'");
			
			while ($attributePrice = shopDbFetchArray($attributesPrices)) {
				switch ($attributePrice['price_prefix']) {
					case '+':	$attributesNetAmount += $attributePrice['options_values_price']; break;
					case '-':	$attributesNetAmount -= $attributePrice['options_values_price']; break;
					default:	break;
				}
			}
		}
		
		if ($finalPrice < $checkAmount) {
			// Produktpreis wurde angepasst -> Attribute werden auf Wert 0 gesetzt und product_price = final_price
			shopDbQuery("UPDATE ".TABLE_ORDERS_PRODUCTS_ATTRIBUTES." SET options_values_price = 0 WHERE orders_products_id = '".shopDbInput($ordersProductsId)."' AND orders_id = '".shopDbInput($orderId)."'");
			
			return $finalPrice;
		} else {
			return $productPrice = $finalPrice - $attributesNetAmount;
		}
	}
}