<?php
/**
 * @version SOFORT Gateway 3.3.0 - $Date: 2013-07-11 17:10:44 +0200 (Thu, 11 Jul 2013) $
 * @author SOFORT AG (integration@sofort.com)
 * @link http://www.sofort.com/
 *
 * Copyright (c) 2012 SOFORT AG
 *
 * Released under the GNU General Public License (Version 2)
 * [http://www.gnu.org/licenses/gpl-2.0.html]
 *
 * $Id: sofort.php 6239 2013-07-11 15:10:44Z boehm $
 */

require_once(DIR_FS_CATALOG.'ext/modules/payment/sofort/helperFunctions.php');

$language = HelperFunctions::getSofortLanguage($_SESSION['language']);
require_once(DIR_FS_CATALOG.DIR_WS_LANGUAGES.$language.'/modules/payment/sofort_general.php');

/**
 * Superclass of oscommerce modules
 */
class sofort {
	//Shop-Vars
	var $form_action_url, $order_status, $icons_available, $enabled, $code, $description, $title, $sort_order;
	//SOFORT-Vars
	var $sofort_js, $title_extern, $sofort, $invoice, $paymentMethod;
	
	function sofort(){
		if (!isset($_SESSION['sofort'])) $_SESSION['sofort'] = array();
		
		$this->form_action_url = DIR_WS_CATALOG . 'ext/modules/payment/sofort/ressources/scripts/redirect.php';
		$this->email_footer = SOFORT_MULTIPAY_TEXT_EMAIL_FOOTER;
		
		$tmpStatusId = (defined('SOFORT_MULTIPAY_TEMP_STATUS_ID') ? HelperFunctions::checkStatusId(SOFORT_MULTIPAY_TEMP_STATUS_ID) : (int)DEFAULT_ORDERS_STATUS_ID);
		
		//dont set a new status? -> then use DEFAULT_ORDERS_STATUS_ID
		if (HelperFunctions::statusIsUnchangedStatus($tmpStatusId)) $tmpStatusId = DEFAULT_ORDER_STATUS_ID;
		
		$this->order_status = HelperFunctions::checkStatusId($tmpStatusId);
		$this->icons_available = '';
		$this->sofort_js = false;
		$this->sofort = '';
		$this->invoice = '';
		//all other class-vars are initialized in child
		
		//first status in history - set if needed (and add to class-vars) @see class order -> function cart()
		//$this->order_status = 'MyStatus';
	}
	
	
	/**
	 * disables paymentmethod, if $order->billingCountry is not allowed to pay with this payment-method (checks the paymentZone)
	 */
	function update_status() {
		global $order;
		
		$constantValue = constant('SOFORT_'.$this->paymentMethod.'_ZONE');
		
		if (($this->enabled == true) && ((int) $constantValue > 0)) {
			$checkFlag = false;
			$checkQuery = tep_db_query("SELECT		zone_id
										FROM		".HelperFunctions::escapeSql(TABLE_ZONES_TO_GEO_ZONES)."
										WHERE		geo_zone_id = '".HelperFunctions::escapeSql($constantValue)."'
										AND			zone_country_id = '".HelperFunctions::escapeSql($order->billing['country']['id'])."'
										ORDER BY	zone_id
			");
			
			while ($check = tep_db_fetch_array($checkQuery)) {
				if ($check['zone_id'] < 1 || $check['zone_id'] == $order->billing['zone_id']) {
					$checkFlag = true;
					break;
				}
			}
			
			if ($checkFlag == false) {
				$this->enabled = false;
			}
		}
	}
	
	
	function javascript_validation() {
		return '';
	}
	
	
	/**
	 * extended in all modules "by sofort"
	 */
	function pre_confirmation_check() {
		global $cartID, $cart;
		
		// We need the cartID
		if (empty($cart->cartID)) $cartID = $cart->cartID = $cart->generate_cart_id();
		if (!tep_session_is_registered('cartID')) tep_session_register('cartID');
	}
	
	
	/**
	 * call with parent::selection() in child
	 */
	function selection() {
		// return false, if modul was not correctly updated
		if (!defined('SOFORT_MULTIPAY_MODULE_VERSION') || strcasecmp(trim(SOFORT_MULTIPAY_MODULE_VERSION), trim(HelperFunctions::getSofortmodulVersion())) != 0) {
			if (is_object($this->sofort)) {
				$this->sofort->logWarning("Paymentmodul was updated by seller but not correct installed. Installed-version in DB does not match file-versions. Please take a look at SOFORT_MULTIPAY_UPDATE_NOTICE or read the manual!");
			} elseif (is_object($this->invoice)) {
				$this->invoice->logWarning("Paymentmodul was updated by seller but not correct installed. Installed-version in DB does not match file-versions. Please take a look at SOFORT_MULTIPAY_UPDATE_NOTICE or read the manual!");
			}
			
			return false;
		}
		
		
		if($this->sofort_js != true){
			echo SOFORT_MULTIPAY_JS_LIBS;
			$this->sofort_js = true;
		}
		
		if (isset($_SESSION['sofort']['apiKeyIsValid'])) {
			return $_SESSION['sofort']['apiKeyIsValid'];
		} else if (!isset($_SESSION['sofort']['apiKeyIsValid'])) {
			$apiTestResult = HelperFunctions::apiKeyIsValid(SOFORT_MULTIPAY_CONFIGURATION_KEY);
			
			if (!$apiTestResult) {
				if (is_object($this->sofort)) {
					$this->sofort->log("Notice: Function apiKeyIsValid() returned false.");
				} elseif (is_object($this->invoice)) {
					$this->invoice->log("Notice: Function apiKeyIsValid() returned false.");
				}
			}
			
			$_SESSION['sofort']['apiKeyIsValid'] = $apiTestResult;
			
			return $apiTestResult;
		}
	}
	
	
	function confirmation($confirmationMessage='', $fields='') {
		global $cartID, $cart_pn_ID, $customer_id, $languages_id, $language, $order, $order_total_modules, $redirect;
		
		if($fields == '') $fields = array(array());
		
		//this will be displayed on confirmation page "payment information"
		$confirmation = array(
				'title' => $confirmationMessage,
				'fields' =>  $fields
		);
		
		//wait for redirect-page
		if($redirect !== true) return $confirmation;
		
		////////////////////////////////////////////////////////////////////////////////////////////////
		//following code inserts the order into shop-db. its called directly before forwarding to pnag//
		////////////////////////////////////////////////////////////////////////////////////////////////
		
		//check, if an old order is saved and must be set "aborted"
		$this->_checkCancelOrder();
		$order_totals = array();
		
		if (is_array($order_total_modules->modules)) {
			reset($order_total_modules->modules);
			
			while (list(, $value) = each($order_total_modules->modules)) {
				$class = substr($value, 0, strrpos($value, '.'));
				
				if ($GLOBALS[$class]->enabled) {
					for ($i=0, $n=sizeof($GLOBALS[$class]->output); $i<$n; $i++) {
						if (tep_not_null($GLOBALS[$class]->output[$i]['title']) && tep_not_null($GLOBALS[$class]->output[$i]['text'])) {
							$order_totals[] = array(
									'code'		 => $GLOBALS[$class]->code,
									'title'		 => $GLOBALS[$class]->output[$i]['title'],
									'text'		 => $GLOBALS[$class]->output[$i]['text'],
									'value'		 => $GLOBALS[$class]->output[$i]['value'],
									'sort_order' => $GLOBALS[$class]->sort_order,
							);
						}
					}
				}
			}
		}
		
		$sql_data_array = array(
				'customers_id'				  => $customer_id,
				'customers_name'			  => $order->customer['firstname'].' '.$order->customer['lastname'],
				'customers_company'			  => $order->customer['company'],
				'customers_street_address'	  => $order->customer['street_address'],
				'customers_suburb'			  => $order->customer['suburb'],
				'customers_city'			  => $order->customer['city'],
				'customers_postcode'		  => $order->customer['postcode'],
				'customers_state'			  => $order->customer['state'],
				'customers_country'			  => $order->customer['country']['title'],
				'customers_telephone'		  => $order->customer['telephone'],
				'customers_email_address'	  => $order->customer['email_address'],
				'customers_address_format_id' => $order->customer['format_id'],
				'delivery_name'				  => $order->delivery['firstname'].' '.$order->delivery['lastname'],
				'delivery_company'			  => $order->delivery['company'],
				'delivery_street_address'	  => $order->delivery['street_address'],
				'delivery_suburb'			  => $order->delivery['suburb'],
				'delivery_city'				  => $order->delivery['city'],
				'delivery_postcode'			  => $order->delivery['postcode'],
				'delivery_state'			  => $order->delivery['state'],
				'delivery_country'			  => $order->delivery['country']['title'],
				'delivery_address_format_id'  => $order->delivery['format_id'],
				'billing_name'				  => $order->billing['firstname'].' '.$order->billing['lastname'],
				'billing_company'			  => $order->billing['company'],
				'billing_street_address'	  => $order->billing['street_address'],
				'billing_suburb'			  => $order->billing['suburb'],
				'billing_city'				  => $order->billing['city'],
				'billing_postcode'			  => $order->billing['postcode'],
				'billing_state'				  => $order->billing['state'],
				'billing_country'			  => $order->billing['country']['title'],
				'billing_address_format_id'	  => $order->billing['format_id'],
				'payment_method'			  => $order->info['payment_method'],
				'cc_type'					  => $order->info['cc_type'],
				'cc_owner'					  => $order->info['cc_owner'],
				'cc_number'					  => $order->info['cc_number'],
				'cc_expires'				  => $order->info['cc_expires'],
				'orders_status'				  => $order->info['order_status'],
				'currency'					  => $order->info['currency'],
				'currency_value'			  => $order->info['currency_value'],
				'date_purchased'			  => 'now()',
		);
		tep_db_perform(TABLE_ORDERS, $sql_data_array);
		$insert_id = tep_db_insert_id();
		
		for ($i=0, $n=sizeof($order_totals); $i<$n; $i++) {
			$sql_data_array = array(
					'orders_id'	 => $insert_id,
					'title'		 => $order_totals[$i]['title'],
					'text'		 => $order_totals[$i]['text'],
					'value'		 => $order_totals[$i]['value'],
					'class'		 => $order_totals[$i]['code'],
					'sort_order' => $order_totals[$i]['sort_order'],
			);
			tep_db_perform(TABLE_ORDERS_TOTAL, $sql_data_array);
		}
		
		for ($i=0, $n=sizeof($order->products); $i<$n; $i++) {
			$sql_data_array = array(
					'orders_id'			=> $insert_id,
					'products_id'		=> tep_get_prid($order->products[$i]['id']),
					'products_model'	=> $order->products[$i]['model'],
					'products_name'		=> $order->products[$i]['name'],
					'products_price'	=> $order->products[$i]['price'],
					'final_price'		=> $order->products[$i]['final_price'],
					'products_tax'		=> $order->products[$i]['tax'],
					'products_quantity' => $order->products[$i]['qty'],
			);
			tep_db_perform(TABLE_ORDERS_PRODUCTS, $sql_data_array);
			$order_products_id = tep_db_insert_id();
			$attributes_exist = '0';
				
			if (isset($order->products[$i]['attributes'])) {
				$attributes_exist = '1';
				
				for ($j=0, $n2=sizeof($order->products[$i]['attributes']); $j<$n2; $j++) {
					if (DOWNLOAD_ENABLED == 'true') {
						$attributes_query = "SELECT		popt.products_options_name		  ,
														poval.products_options_values_name,
														pa.options_values_price			  ,
														pa.price_prefix					  ,
														pad.products_attributes_maxdays	  ,
														pad.products_attributes_maxcount  ,
														pad.products_attributes_filename
											FROM		".TABLE_PRODUCTS_OPTIONS." popt		   ,
														".TABLE_PRODUCTS_OPTIONS_VALUES." poval,
														".TABLE_PRODUCTS_ATTRIBUTES." pa
											LEFT JOIN	".TABLE_PRODUCTS_ATTRIBUTES_DOWNLOAD." pad
											ON			pa.products_attributes_id=pad.products_attributes_id
											WHERE pa.products_id = '" . HelperFunctions::escapeSql($order->products[$i]['id']) . "'
											AND			pa.options_id = '".HelperFunctions::escapeSql($order->products[$i]['attributes'][$j]['option_id'])."'
											AND			pa.options_id = popt.products_options_id
											AND			pa.options_values_id = '".HelperFunctions::escapeSql($order->products[$i]['attributes'][$j]['value_id'])."'
											AND			pa.options_values_id = poval.products_options_values_id
											AND			popt.language_id = '".HelperFunctions::escapeSql($languages_id)."'
											AND			poval.language_id = '".HelperFunctions::escapeSql($languages_id)."'
						";
						$attributes = tep_db_query($attributes_query);
					} else {
						$attributes = tep_db_query("SELECT	popt.products_options_name		  ,
															poval.products_options_values_name,
															pa.options_values_price			  ,
															pa.price_prefix
													FROM	".TABLE_PRODUCTS_OPTIONS." popt		   ,
															".TABLE_PRODUCTS_OPTIONS_VALUES." poval,
															".TABLE_PRODUCTS_ATTRIBUTES." pa
													WHERE	pa.products_id = '".HelperFunctions::escapeSql($order->products[$i]['id'])."'
													AND		pa.options_id = '".HelperFunctions::escapeSql($order->products[$i]['attributes'][$j]['option_id'])."'
													AND		pa.options_id = popt.products_options_id
													AND		pa.options_values_id = '".HelperFunctions::escapeSql($order->products[$i]['attributes'][$j]['value_id'])."'
													AND		pa.options_values_id = poval.products_options_values_id
													AND		popt.language_id = '".HelperFunctions::escapeSql($languages_id)."'
													AND		poval.language_id = '".HelperFunctions::escapeSql($languages_id)."'
						");
					}
					
					$attributes_values = tep_db_fetch_array($attributes);
					$sql_data_array = array(
							'orders_id'				  => $insert_id,
							'orders_products_id'	  => $order_products_id,
							'products_options'		  => $attributes_values['products_options_name'],
							'products_options_values' => $attributes_values['products_options_values_name'],
							'options_values_price'	  => $attributes_values['options_values_price'],
							'price_prefix'			  => $attributes_values['price_prefix'],
					);
					tep_db_perform(TABLE_ORDERS_PRODUCTS_ATTRIBUTES, $sql_data_array);
					
					if ((DOWNLOAD_ENABLED == 'true') && isset($attributes_values['products_attributes_filename']) && tep_not_null($attributes_values['products_attributes_filename'])) {
						$sql_data_array = array(
								'orders_id'				   => $insert_id,
								'orders_products_id'	   => $order_products_id,
								'orders_products_filename' => $attributes_values['products_attributes_filename'],
								'download_maxdays'		   => $attributes_values['products_attributes_maxdays'],
								'download_count'		   => $attributes_values['products_attributes_maxcount'],
						);
						tep_db_perform(TABLE_ORDERS_PRODUCTS_DOWNLOAD, $sql_data_array);
					}
				}
			}
		}
		
		$cart_pn_ID = $cartID.'-'.$insert_id;
		tep_session_register('cart_pn_ID');
		
		return true;
	}
	
	
	function process_button() {
		global $cart, $cart_pn_ID, $currencies, $currency, $customer_id, $insert_id, $order, $order_totals, $redirect;
		
		if ($redirect === true) {
			$orderId = substr($cart_pn_ID, strpos($cart_pn_ID, '-')+1);
			$reasons = $this->_getReasons($this->paymentMethod, $customerId, $order, $orderId);
			$paymentSecret = md5(mt_rand().microtime());
			$userVariable_0 = tep_output_string($orderId);
			$userVariable_1 = tep_output_string($customer_id);
			$userVariable_2 = tep_output_string($cart->cartID);
			$userVariable_3 = $paymentSecret;
			$amount = number_format($order->info['total'] * $currencies->get_value($currency), 2, '.','');
			
			$successUrl = tep_href_link(FILENAME_CHECKOUT_PROCESS, '', 'SSL', true, false);
			$successUrl = HelperFunctions::cleanUrlParameter($successUrl);
			
			$cancelUrl = tep_href_link(FILENAME_CHECKOUT_PAYMENT, 'payment_error='.$this->code.'&tid=-TRANSACTION-', 'SSL', true, false);
			$cancelUrl = HelperFunctions::cleanUrlParameter($cancelUrl);
			
			$notificationUrl = tep_href_link('ext/modules/payment/sofort/callback.php', 'paymentSecret='.$paymentSecret.'&action=multipay', 'SSL', true, false);
			$notificationUrl = HelperFunctions::cleanUrlParameter($notificationUrl);
			
			$this->sofort->setAmount($amount, $currency);
			$this->sofort->setReason(HelperFunctions::convertEncoding($reasons[0],3), HelperFunctions::convertEncoding($reasons[1],3));
			$this->sofort->setSuccessUrl(HelperFunctions::convertEncoding($successUrl,3));
			$this->sofort->setAbortUrl(HelperFunctions::convertEncoding($cancelUrl,3));
			$this->sofort->setTimeoutUrl(HelperFunctions::convertEncoding($cancelUrl,3));
			$this->sofort->setNotificationUrl(HelperFunctions::convertEncoding($notificationUrl,3));
			$this->sofort->addUserVariable(HelperFunctions::convertEncoding($userVariable_0,3));
			$this->sofort->addUserVariable(HelperFunctions::convertEncoding($userVariable_1,3));
			$this->sofort->addUserVariable(HelperFunctions::convertEncoding($userVariable_2,3));
			$this->sofort->addUserVariable(HelperFunctions::convertEncoding($userVariable_3,3));
			$this->sofort->setEmailCustomer(HelperFunctions::convertEncoding($order->customer['email_address'],3));
			$this->sofort->setPhoneNumberCustomer(HelperFunctions::convertEncoding($order->customer['telephone'],3));
			
			//set special settings for every payment-method
			switch($this->paymentMethod) {
				case 'SU' :
					$this->sofort->setSofortueberweisung($amount);
					// see if customer protection is enabled, set it as parameter to sofortlib
					$this->sofort->setSofortueberweisungCustomerprotection(SOFORT_SU_KS_STATUS == 'True');
					break;
				case 'LS' :
					$this->sofort->setLastschrift();
					$this->sofort->setSenderAccount(HelperFunctions::convertEncoding($_SESSION['sofort']['ls_bank_code'],3),
													HelperFunctions::convertEncoding($_SESSION['sofort']['ls_account_number'],3),
													HelperFunctions::convertEncoding($_SESSION['sofort']['ls_sender_holder'],3));
						
					$billingSalutation = $this->_getGenderFromAddressBook($order->billing['firstname'], $order->billing['lastname'], $order->billing['company'], $order->billing['street_address'], $order->billing['postcode'],
																		  $order->billing['city'], $order->billing['country_id'], $order->billing['zone_id']);
						
					//split street and number
					if (!preg_match('#(.+)[ .](.+)#i', trim($order->billing['street_address']), $streetparts)) {
						$streetparts = array();
						$streetparts[1] = trim($order->billing['street_address']);
						$streetparts[2] = '';
					}
						
					//currently $order->billing['suburb'] (german: "Adresszusatz") is not supported by sofortLib
						
					$this->sofort->setLastschriftAddress(HelperFunctions::convertEncoding($order->billing['firstname'],3), HelperFunctions::convertEncoding($order->billing['lastname'],3), HelperFunctions::convertEncoding($streetparts[1],3),
														 HelperFunctions::convertEncoding($streetparts[2],3), HelperFunctions::convertEncoding($order->billing['postcode'],3), HelperFunctions::convertEncoding($order->billing['city'],3),
														 HelperFunctions::convertEncoding($billingSalutation,3), HelperFunctions::convertEncoding($order->billing['country']['iso_code_2'],3));
					break;
			}
			
			//send all data against the SOFORT-API
			$this->sofort->sendRequest();
				
			if ($this->sofort->isError()) {
				$this->sofort->logWarning("API-Call returned false. Redirect to cancel-URL. API-Errors: ".print_r($this->sofort->getErrors(), true)." Time: ".date("d.m.Y, G:i:s"));
				tep_redirect(HelperFunctions::getCancelUrl($this->code, $this->sofort->getErrors()));
			} else {
				$url = $this->sofort->getPaymentUrl();
				$transactionId = $this->sofort->getTransactionId();
				
				//seller and customer comment
				$time = date("d.m.Y, G:i:s");
				
				//set temp-status (only table orders_history, not table orders)
				$tmpStatusId = SOFORT_MULTIPAY_TEMP_STATUS_ID;
				
				//dont set a new status? -> then use the last order status
				if (HelperFunctions::statusIsUnchangedStatus($tmpStatusId)) $tmpStatusId = HelperFunctions::getLastOrderStatus($orderId);
				
				$tmpStatusId = HelperFunctions::checkStatusId($tmpStatusId);
				
				//comment only for buyer and seller
				$tmpComment = constant('SOFORT_'.$this->paymentMethod.'_TMP_COMMENT').' '.SOFORT_TRANSLATE_TIME.': '.$time;
				HelperFunctions::insertHistoryEntry((int) $orderId, $tmpStatusId, $tmpComment);
				
				//comment only for seller
				$tmpCommentSeller = constant('SOFORT_'.$this->paymentMethod.'_TMP_COMMENT_SELLER').' '.SOFORT_TRANSLATE_TIME.': '.$time;
				HelperFunctions::insertHistoryEntry((int) $orderId, -1, $tmpCommentSeller, 0);
				
				//save transaction-id
				HelperFunctions::insertHistoryEntry((int) $orderId, -1, $transactionId, 0);
				
				//save all important data in our sofort-tables
				$sofortOrderId = HelperFunctions::insertSofortOrder($orderId, $paymentSecret, $transactionId, $this->paymentMethod);
				
				if (!$sofortOrderId) {
					$this->sofort->logWarning("Warning: Saving of orderdetails in table sofort_orders failed. Function insertSofortOrder() returned false.
							Given params: OrderId: $orderId, PaymentSecret: $paymentSecret, TransId: $transactionId, PaymentMethod:".$this->paymentMethod);
				}
				
				$_SESSION['sofort']['sofort_payment_url'] = $url;
				$_SESSION['sofort']['sofort_payment_method'] = $this->code;
			}
			
			for ($i=0, $n=sizeof($order->products); $i<$n; $i++) {
				// Stock Update - Joao Correia
				if (STOCK_LIMITED == 'true') {
					if (DOWNLOAD_ENABLED == 'true') {
						$stock_query_raw = "SELECT		products_quantity,
														pad.products_attributes_filename
											FROM		".TABLE_PRODUCTS." p
											LEFT JOIN	".TABLE_PRODUCTS_ATTRIBUTES." pa
											ON			p.products_id=pa.products_id
											LEFT JOIN	".TABLE_PRODUCTS_ATTRIBUTES_DOWNLOAD." pad
											ON			pa.products_attributes_id=pad.products_attributes_id
											WHERE		p.products_id = '".tep_get_prid(HelperFunctions::escapeSql($order->products[$i]['id']))."'
						";
							
						// Will work with only one option for downloadable products
						// otherwise, we have to build the query dynamically with a loop
						$products_attributes = $order->products[$i]['attributes'];
							
						if (is_array($products_attributes)) {
							$stock_query_raw .= " AND pa.options_id = '".HelperFunctions::escapeSql($products_attributes[0]['option_id'])."' AND pa.options_values_id = '".HelperFunctions::escapeSql($products_attributes[0]['value_id'])."'";
						}
							
						$stock_query = tep_db_query($stock_query_raw);
					} else {
						$stock_query = tep_db_query("SELECT products_quantity FROM ".TABLE_PRODUCTS." WHERE products_id = '".tep_get_prid(HelperFunctions::escapeSql($order->products[$i]['id']))."'");
					}
					if (tep_db_num_rows($stock_query) > 0) {
						$stock_values = tep_db_fetch_array($stock_query);
						// do not decrement quantities if products_attributes_filename exists
						$stock_left = (DOWNLOAD_ENABLED != 'true' || !$stock_values['products_attributes_filename']) ? $stock_values['products_quantity'] - $order->products[$i]['qty'] : $stock_values['products_quantity'];
						tep_db_query("UPDATE ".TABLE_PRODUCTS." SET products_quantity = '".HelperFunctions::escapeSql($stock_left)."' WHERE products_id = '".tep_get_prid(HelperFunctions::escapeSql($order->products[$i]['id']))."'");
							
						if ( ($stock_left < 1) && (STOCK_ALLOW_CHECKOUT == 'false') ) {
							tep_db_query("UPDATE ".TABLE_PRODUCTS." SET products_status = '0' WHERE products_id = '".tep_get_prid(HelperFunctions::escapeSql($order->products[$i]['id']))."'");
						}
					}
				}
			}
			
			return $url;
		} else {
			return '';
		}
	}
	
	
	function before_process() {
		global $customer_id, $order, $order_totals, $sendto, $billto, $insert_id, $languages_id, $payment, $currencies, $currency, $cart, $cart_pn_ID, $$payment;
		
		//fix bug in < osc 2.2 RC2
		if(!is_array($order_totals)){
			include(DIR_WS_CLASSES.'order_total.php');
			$order_total_modules = new order_total;
			$order_totals = $order_total_modules->process();
		}
		
		$order_id = substr($cart_pn_ID, strpos($cart_pn_ID, '-')+1);
		$tmpStatusId = SOFORT_MULTIPAY_TEMP_STATUS_ID;
		
		if(!empty($order->info['comments'])){
			$sql_data_array = array(
					'orders_id'			=> $order_id,
					'orders_status_id'	=> $tmpStatusId,
					'date_added'		=> 'now()',
					'customer_notified'	=> (SEND_EMAILS == 'true') ? '1' : '0',
					'comments'			=> $order->info['comments']);
			tep_db_perform(TABLE_ORDERS_STATUS_HISTORY, $sql_data_array);
		}
		
		$order_totals_table_top = '<table border="0" cellpadding="5" cellspacing="0">';
		$order_totals_cell_top = '<tr><td width="280" style="font-size: 12px">';
		$order_totals_cell_mid = '</td><td style="font-size: 12px" align="right">';
		$order_totals_cell_end = '</td></tr>';
		$order_totals_table_end = '</table>';
		
		// initialized for the email confirmation
		$products_ordered = (EMAIL_USE_HTML == 'true') ? $order_totals_table_top : '';
		
		$subtotal = 0;
		$total_tax = 0;
		
		for ($i=0, $n=sizeof($order->products); $i<$n; $i++) {
			// Update products_ordered (for bestsellers list)
			tep_db_query("UPDATE ".TABLE_PRODUCTS." SET products_ordered = products_ordered + ".sprintf('%d', $order->products[$i]['qty'])." WHERE products_id = '".tep_get_prid(HelperFunctions::escapeSql($order->products[$i]['id']))."'");
				
			//insert customer chosen option to order
			$attributes_exist = '0';
			$products_ordered_attributes = '';
				
			if (isset($order->products[$i]['attributes'])) {
				$attributes_exist = '1';
				
				for ($j=0, $n2=sizeof($order->products[$i]['attributes']); $j<$n2; $j++) {
					if (DOWNLOAD_ENABLED == 'true') {
						$attributes_query = "SELECT		popt.products_options_name		  ,
														poval.products_options_values_name,
														pa.options_values_price			  ,
														pa.price_prefix					  ,
														pad.products_attributes_maxdays	  ,
														pad.products_attributes_maxcount  ,
														pad.products_attributes_filename
											FROM		".TABLE_PRODUCTS_OPTIONS." popt		   ,
														".TABLE_PRODUCTS_OPTIONS_VALUES." poval,
														".TABLE_PRODUCTS_ATTRIBUTES." pa
											LEFT JOIN	".TABLE_PRODUCTS_ATTRIBUTES_DOWNLOAD." pad
											ON			pa.products_attributes_id=pad.products_attributes_id
											WHERE		pa.products_id = '".HelperFunctions::escapeSql($order->products[$i]['id'])."'
											AND		pa.options_id = '".HelperFunctions::escapeSql($order->products[$i]['attributes'][$j]['option_id'])."'
											AND		pa.options_id = popt.products_options_id
											AND		pa.options_values_id = '".HelperFunctions::escapeSql($order->products[$i]['attributes'][$j]['value_id'])."'
											AND		pa.options_values_id = poval.products_options_values_id
											AND		popt.language_id = '".HelperFunctions::escapeSql($languages_id)."'
											AND		poval.language_id = '".HelperFunctions::escapeSql($languages_id)."'
						";
						$attributes = tep_db_query($attributes_query);
					} else {
						$attributes = tep_db_query("SELECT	popt.products_options_name,
															poval.products_options_values_name,
															pa.options_values_price,
															pa.price_prefix
													FROM	".TABLE_PRODUCTS_OPTIONS." popt,
															".TABLE_PRODUCTS_OPTIONS_VALUES." poval,
															".TABLE_PRODUCTS_ATTRIBUTES." pa
													WHERE 	pa.products_id = '".HelperFunctions::escapeSql($order->products[$i]['id'])."'
													AND		pa.options_id = '".HelperFunctions::escapeSql($order->products[$i]['attributes'][$j]['option_id'])."'
													AND		pa.options_id = popt.products_options_id and pa.options_values_id = '".HelperFunctions::escapeSql($order->products[$i]['attributes'][$j]['value_id'])."'
													AND 	pa.options_values_id = poval.products_options_values_id and popt.language_id = '".HelperFunctions::escapeSql($languages_id)."'
													AND 	poval.language_id = '".HelperFunctions::escapeSql($languages_id)."'
								");
					}
						
					$attributes_values = tep_db_fetch_array($attributes);
					$products_ordered_attributes .= "\n\t".$attributes_values['products_options_name'].' '.$attributes_values['products_options_values_name'];
				}
			}
				
			//------insert customer choosen option eof ----
			$total_weight += ($order->products[$i]['qty'] * $order->products[$i]['weight']);
			$total_tax += tep_calculate_tax($total_products_price, $products_tax) * $order->products[$i]['qty'];
			$total_cost += $total_products_price;
				
			if (EMAIL_USE_HTML == 'true'){
				$products_ordered .= $order_totals_cell_top;
				$products_ordered .= (int)$order->products[$i]['qty'].' x '.$this->cleanString($order->products[$i]['name']);
				$products_ordered .= ' ('.$this->cleanString($order->products[$i]['model']).') = ';
				$prodcuts_ordered .= $order_totals_cell_mid;
				$products_ordered .= $currencies->display_price($this->cleanString($order->products[$i]['final_price']),$this->cleanString($order->products[$i]['tax']),(int)$order->products[$i]['qty']).$products_ordered_attributes;
				$products_ordered .= $order_totals_cell_end;
			} else {
				$products_ordered .= (int)$order->products[$i]['qty'].' x '.$this->cleanString($order->products[$i]['name']);
				$products_ordered .= ' ('.$this->cleanString($order->products[$i]['model']).') = ';
				$products_ordered .= $currencies->display_price($this->cleanString($order->products[$i]['final_price']),$this->cleanString($order->products[$i]['tax']),(int)$order->products[$i]['qty']).$products_ordered_attributes."\n";
			}
		}
		
		if (EMAIL_USE_HTML == 'true'){
			$products_ordered .= $order_totals_table_end;
			$text = preg_replace('#([\n\t])|([ ]+)#', ' ', SOFORT_MULTIPAY_EMAIL_HTML_TEXT);
		} else {
			$text = SOFORT_MULTIPAY_EMAIL_TEXT;
		}
		
		$text = str_replace('{{ORDER_ID}}', $order_id, $text);
		$text = str_replace('{{INVOICE_URL}}', tep_href_link(FILENAME_ACCOUNT_HISTORY_INFO, 'order_id=' . $order_id, 'SSL', false), $text);
		$text = str_replace('{{DATE_ORDERED}}', strftime(DATE_FORMAT_LONG), $text);
		$text = ($order->info['comments']) ? str_replace('{{CUSTOMER_COMMENT}}', tep_db_output($order->info['comments']), $text) : str_replace('{{CUSTOMER_COMMENT}}', '', $text);
		$text = str_replace('{{Item_List}}', $products_ordered, $text);
		
		if (EMAIL_USE_HTML == 'true'){
			$list_total = $order_totals_table_top;
				
			for ($i=0, $n=sizeof($order_totals); $i<$n; $i++) {
				$list_total .= $order_totals_cell_top.$this->cleanString($order_totals[$i]['title']).$order_totals_cell_mid.$this->cleanString($order_totals[$i]['text']).$order_totals_cell_end;
			}
				
			$list_total .= $order_totals_table_end;
		} else {
			for ($i=0, $n=sizeof($order_totals); $i<$n; $i++) {
				$list_total .= $this->cleanString($order_totals[$i]['title']).' '.$this->cleanString($order_totals[$i]['text'])."\n";
			}
		}
		
		$text = str_replace('{{List_Total}}', $list_total, $text);
		
		if ($order->content_type != 'virtual') {
			$text = str_replace('{{DELIVERY_ADRESS}}', tep_address_label($customer_id, $sendto, 0, '', "\n"), $text);
		} elseif ($order->content_type == 'virtual') {
			if ((DOWNLOAD_ENABLED == 'true') && isset($attributes_values['products_attributes_filename']) && tep_not_null($attributes_values['products_attributes_filename'])) {
				$text = str_replace('{{DELIVERY_ADRESS}}', tep_href_link(FILENAME_ACCOUNT_HISTORY_INFO, 'order_id=' . $order_id, 'SSL', false), $text);
			} else {
				$text = str_replace('{{DELIVERY_ADRESS}}', '', $text);
			}
		} else {
			$text = str_replace('{{DELIVERY_ADRESS}}', '', $text);
		}
		
		$text = str_replace('{{BILLING_ADRESS}}', tep_address_label($customer_id, $billto, 0, '', "\n"), $text);
		
		if (is_object($$payment)) {
			$payment_class = $$payment;
			$payment_text = $payment_class->public_title;
			
			$text = str_replace('{{Payment_Modul_Text}}', $payment_text, $text);
			$payment_email_footer = ''.$payment_class->email_footer;
			$text = str_replace('{{Payment_Modul_Text_Footer}}', $payment_email_footer, $text);
		}
		
		//replace all {{STORE_*}} {{OWNER_*}} {{EMAIL_*}} with defines
		//STORE_NAME STORE_NAME_ADDRESS OWNER_BANK_FA OWNER_BANK_TAX_NUMBER OWNER_BANK_UST_NUMBER OWNER_BANK_NAME OWNER_BANK_ACCOUNT STORE_OWNER_BLZ OWNER_BANK OWNER_BANK_SWIFT OWNER_BANK_IBAN
		$text = preg_replace("#\{\{((STORE_|OWNER_|EMAIL_)[A-Z0-9_]+)\}\}#e", '$1', $text);
		$email_order = $text;
		$email_subject = SOFORT_MULTIPAY_TEXT_EMAIL_SUBJECT;
		$email_subject = preg_replace("#\{\{((STORE_|OWNER_|EMAIL_)[A-Z0-9_]+)\}\}#e", '$1', $email_subject);
		$email_subject = str_replace('{{ORDER_ID}}', $order_id, $email_subject);
		$email_subject = str_replace('{{DATE_ORDERED}}', strftime(DATE_FORMAT_LONG), $email_subject);
		
		if(SEND_EMAILS == 'true' && $order_id) {
			tep_mail($this->cleanString($order->customer['firstname']).' '.$this->cleanString($order->customer['lastname']), $this->cleanString($order->customer['email_address']),$this->cleanString($email_subject),$email_order,STORE_OWNER, STORE_OWNER_EMAIL_ADDRESS);
			
			//set the customer-notified-flag
			tep_db_query("UPDATE ".TABLE_ORDERS_STATUS_HISTORY." SET customer_notified='1' WHERE orders_id='".(int)$order_id."' ORDER BY orders_status_history_id ASC LIMIT 1");
		}
		
		// send emails to other people
		if (SEND_EXTRA_ORDER_EMAILS_TO != '') {
			tep_mail('', SEND_EXTRA_ORDER_EMAILS_TO, $email_subject , $email_order, STORE_OWNER, STORE_OWNER_EMAIL_ADDRESS);
		}
		
		$insert_id = $order_id;
		//different circumstances can cause this part to be called, before the first notification arrives.
		//to prevent the loss of information which could occur when the first notification never arrives the
		//first status and comments will be set here (if not already happened)
		$this->_insertPaidStatus();
		
		$this->after_process();
		$cart->reset(true);
		
		// unregister session variables used during checkout
		tep_session_unregister('sendto');
		tep_session_unregister('billto');
		tep_session_unregister('shipping');
		tep_session_unregister('payment');
		tep_session_unregister('comments');
		tep_session_unregister('cart_pn_ID');
		
		if ($_SESSION['sofort']['ls_sender_holder'] ) 	  unset ($_SESSION['sofort']['ls_sender_holder']);
		if ($_SESSION['sofort']['ls_account_number'] ) 	  unset ($_SESSION['sofort']['ls_account_number']);
		if ($_SESSION['sofort']['ls_bank_code'] ) 		  unset ($_SESSION['sofort']['ls_bank_code']);
		if ($_SESSION['sofort']['sofort_conditions_sr'] ) unset ($_SESSION['sofort']['sofort_conditions_sr']);
		if ($_SESSION['sofort']['sofort_conditions_ls'] ) unset ($_SESSION['sofort']['sofort_conditions_ls']);
		
		tep_redirect(tep_href_link(FILENAME_CHECKOUT_SUCCESS, '', 'SSL'));
	}
	
	
	/**
	 * called by core after successful payment
	 */
	function after_process() {
		return false;
	}
	
	
	/**
	 * check if there is an error with this paymentMethod - called in checkout_payment.php
	 * @return array with title and description of the found error(s)
	 */
	function get_error () {
		$this->_checkCancelOrder();
		
		if (!isset($_GET['payment_error']) || $_GET['payment_error'] != $this->code) return false;
		
		//there is an error with this paymentMethod...
		$this->enabled = false;
		$errormsgArray = array();
		
		if (isset($_GET['error_codes'])) {
			$langConstantExist = true;
			$errorCodes = array_unique (explode(',', $_GET['error_codes']));
			
			foreach ($errorCodes as $errorCode) {
				$errorCode = trim($errorCode);
				$code = substr($errorCode, 0, strpos($errorCode, '.'));
				
				if ($code === false || empty($code)) $code = $this->cleanString($errorCode);
				
				if (defined ('SOFORT_MULTIPAY_XML_FAULT_' . strtoupper($errorCode) ) ) {
					$errormsgArray[] = constant('SOFORT_MULTIPAY_XML_FAULT_'.strtoupper($errorCode) ).' ('.$code.')';
				} elseif (defined('SOFORT_MULTIPAY_XML_FAULT_' . $code) ) {
					$errormsgArray[] = constant('SOFORT_MULTIPAY_XML_FAULT_'.strtoupper($code)).' ('.$code.')';
				} else {
					$langConstantExist = false;
				}
				
				//show this paymentMethod again to buyer with the following error-codes
				$dontDisableCodes = array('10000', '10001', '10002');
				
				if(in_array($code, $dontDisableCodes) ) $this->enabled = true;
			}
			
			if (!$errormsgArray && $langConstantExist == false) $errormsgArray[] = constant('SOFORT_MULTIPAY_XML_FAULT_'.$this->paymentMethod);
		} else {
			$errormsgArray[] = constant('SOFORT_'.$this->paymentMethod.'_TEXT_ERROR_MESSAGE');
		}
		
		if (!$errormsgArray) $errormsgArray[] = constant('SOFORT_MULTIPAY_XML_FAULT_'.$this->paymentMethod);
		
		$errormsgArray = array_unique($errormsgArray);
		$title = SOFORT_MULTIPAY_TEXT_ERROR_HEADING;
		
		return array(
				'title' => $title,
				'error' => implode(' ', $errormsgArray),
		);
	}
	
	
	/**
	 * Module is installed?
	 */
	function check() {
		if (!isset($this->_check)) {
			$constantName = 'SOFORT_'.$this->paymentMethod.'_STATUS';
			$check_query = tep_db_query("SELECT configuration_value FROM " . TABLE_CONFIGURATION . " WHERE configuration_key = '".HelperFunctions::escapeSql($constantName)."'");
			$this->_check = tep_db_num_rows($check_query);
		}
		
		if ($this->_check) {
			$this->_installSofortOrdersTable();
			$this->_installSofortOrdersNotificationTable();
			$this->_installSofortProductsTable();
		}
		
		return $this->_check;
	}


	/**
	 * abstract method
	 * must be implemented in concrete implementation
	 */
	function install() {
		//following defined constants are global(!) - that means, they are set only once and are used by most/all payment-methods
		if(!defined('SOFORT_MULTIPAY_CONFIGURATION_KEY')) {
			$sofortStatuses = $this->_insertAndReturnSofortStatus();
			$tempStatus = (isset($sofortStatuses['temp'])&& !empty($sofortStatuses['temp']))? $sofortStatuses['temp'] : '';
			$abortedStatus = (isset($sofortStatuses['aborted'])&& !empty($sofortStatuses['aborted']))? $sofortStatuses['aborted'] : '';
			$unchangedStatus = (isset($sofortStatuses['unchanged'])&& !empty($sofortStatuses['unchanged']))? $sofortStatuses['unchanged'] : '';
			//$checkStatus = (isset($sofortStatuses['check'])&& !empty($sofortStatuses['check']))? $sofortStatuses['check'] : '';
			
			tep_db_query("INSERT INTO ".TABLE_CONFIGURATION." (configuration_title, configuration_description, configuration_key, configuration_value,  configuration_group_id, sort_order, date_added) VALUES ('".HelperFunctions::escapeSql(SOFORT_MULTIPAY_CONFIGURATION_KEY_TITLE)."', '".HelperFunctions::escapeSql(SOFORT_MULTIPAY_CONFIGURATION_KEY_DESC)."', 'SOFORT_MULTIPAY_CONFIGURATION_KEY', '',  '6', '2', now())");
			tep_db_query("INSERT INTO ".TABLE_CONFIGURATION." (configuration_title, configuration_description, configuration_key, configuration_value,  configuration_group_id, sort_order, set_function, date_added) VALUES ('".HelperFunctions::escapeSql(SOFORT_MULTIPAY_AUTH_TITLE)."', '".HelperFunctions::escapeSql(SOFORT_MULTIPAY_AUTH_DESC)."', 'SOFORT_MULTIPAY_AUTH', '".HelperFunctions::escapeSql(SOFORT_KEYTEST_DEFAULT)."',  '6', '3', 'tep_cfg_select_option(array(),', now())");  //hide the input-field with an empty <select>
			tep_db_query("INSERT INTO ".TABLE_CONFIGURATION." (configuration_title, configuration_description, configuration_key, configuration_value,  configuration_group_id, sort_order, set_function, date_added) VALUES ('".HelperFunctions::escapeSql(SOFORT_MULTIPAY_REASON_1_TITLE)."', '".HelperFunctions::escapeSql(SOFORT_MULTIPAY_REASON_1_DESC)."', 'SOFORT_MULTIPAY_REASON_1', '-TRANSACTION-',  '6', '8', 'tep_cfg_select_option(array(\'Kd-Nr. {{customer_id}}\',\'-TRANSACTION-\'), ', now())");
			tep_db_query("INSERT INTO ".TABLE_CONFIGURATION." (configuration_title, configuration_description, configuration_key, configuration_value,  configuration_group_id, sort_order, date_added) VALUES ('".HelperFunctions::escapeSql(SOFORT_MULTIPAY_REASON_2_TITLE)."', '".HelperFunctions::escapeSql(SOFORT_MULTIPAY_REASON_2_DESC)."', 'SOFORT_MULTIPAY_REASON_2', '" . HelperFunctions::escapeSql(STORE_NAME) . "', '6', '9', now())");
			tep_db_query("INSERT INTO ".TABLE_CONFIGURATION." (configuration_title, configuration_description, configuration_key, configuration_value,  configuration_group_id, sort_order, set_function, date_added) VALUES ('".HelperFunctions::escapeSql(SOFORT_MULTIPAY_IMAGE_TITLE)."', '".HelperFunctions::escapeSql(SOFORT_MULTIPAY_IMAGE_DESC)."', 'SOFORT_MULTIPAY_IMAGE', 'Infographic',  '6', '7', 'tep_cfg_select_option(array(\'Infographic\',\'Logo & Text\'), ', now())");
			tep_db_query("INSERT INTO ".TABLE_CONFIGURATION." (configuration_title, configuration_description, configuration_key, configuration_value,  configuration_group_id, sort_order, set_function, date_added) VALUES ('".HelperFunctions::escapeSql(SOFORT_MULTIPAY_PROF_SETTINGS_TITLE)."', '".HelperFunctions::escapeSql(SOFORT_MULTIPAY_PROF_SETTINGS_DESC)."', 'SOFORT_MULTIPAY_PROF_SETTINGS', '',  '6', '20', 'tep_cfg_select_option(array(),', now())");  //hide the input-field with an empty <select>
			tep_db_query("INSERT INTO ".TABLE_CONFIGURATION." (configuration_title, configuration_description, configuration_key, configuration_value,  configuration_group_id, sort_order, set_function, use_function, date_added) VALUES ('".HelperFunctions::escapeSql(SOFORT_MULTIPAY_TEMP_STATUS_ID_TITLE)."', '".HelperFunctions::escapeSql(SOFORT_MULTIPAY_TEMP_STATUS_ID_DESC)."', 'SOFORT_MULTIPAY_TEMP_STATUS_ID', '".HelperFunctions::escapeSql($tempStatus)."',  '6', '35', 'tep_cfg_pull_down_order_statuses(', 'tep_get_order_status_name', now())");
			tep_db_query("INSERT INTO ".TABLE_CONFIGURATION." (configuration_title, configuration_description, configuration_key, configuration_value,  configuration_group_id, sort_order, set_function, use_function, date_added) VALUES ('".HelperFunctions::escapeSql(SOFORT_MULTIPAY_ABORTED_STATUS_ID_TITLE)."', '".HelperFunctions::escapeSql(SOFORT_MULTIPAY_ABORTED_STATUS_ID_DESC)."', 'SOFORT_MULTIPAY_ABORTED_STATUS_ID', '".HelperFunctions::escapeSql($abortedStatus)."',  '6', '35', 'tep_cfg_pull_down_order_statuses(', 'tep_get_order_status_name', now())");
			tep_db_query("INSERT INTO ".TABLE_CONFIGURATION." (configuration_title, configuration_description, configuration_key, configuration_value,  configuration_group_id, sort_order, set_function, date_added) VALUES ('".HelperFunctions::escapeSql(SOFORT_MULTIPAY_LOG_ENABLED_TITLE)."', '".HelperFunctions::escapeSql(SOFORT_MULTIPAY_LOG_ENABLED_DESC)."', 'SOFORT_MULTIPAY_LOG_ENABLED', 'False', '6', '28', 'tep_cfg_select_option(array(\'True\', \'False\'), ', now())");
			
			//only used while setting status - unchangeable and not shown in module-configuration
			tep_db_query("INSERT INTO ".TABLE_CONFIGURATION." (configuration_title, configuration_description, configuration_key, configuration_value,  configuration_group_id, sort_order, date_added) VALUES ('unimportant', 'unimportant', 'SOFORT_MULTIPAY_UNCHANGED_STATUS_ID', '".HelperFunctions::escapeSql($unchangedStatus)."',  '6', '35', now())");
			
			$moduleVersion = HelperFunctions::getSofortmodulVersion();
			tep_db_query("INSERT INTO ".TABLE_CONFIGURATION." (configuration_title, configuration_description, configuration_key, configuration_value,  configuration_group_id, sort_order, date_added) VALUES ('unimportant', 'unimportant', 'SOFORT_MULTIPAY_MODULE_VERSION', '".HelperFunctions::escapeSql($moduleVersion)."',  '6', '2', now())");
			
			$this->_installSofortOrdersTable();
			$this->_installSofortOrdersNotificationTable();
			$this->_installSofortProductsTable();
		}
		return true;
	}
	
	
	/**
	 * abstract method
	 * must be implemented in concrete implementation
	 */
	function remove() {
		//if this is the last deletion of a multipay-paymentmethod --> we also delete the configurationKey and other values
		$check_query = tep_db_query("SELECT * FROM ".TABLE_CONFIGURATION." WHERE configuration_key like 'SOFORT_%_STATUS' AND configuration_key NOT LIKE '%_IDEAL_%'");
		
		if (tep_db_num_rows ($check_query) === 0 ) {
			tep_db_query("DELETE FROM ".TABLE_CONFIGURATION." WHERE configuration_key LIKE 'SOFORT_MULTIPAY_%' AND configuration_key NOT LIKE '%_IDEAL_%'");
			
			//We don't want to delete data, but we could in some cases...
			//Notice: Sofort-orderhistory will also be deleted and cannot be shown anymore
			//$this->_removeSofortOrdersTable();
			//$this->_removeSofortOrdersNotificationTable();
			//$this->_removeSofortProductsTable();
		}
	}
	
	
	function keys() {
		//unset, important for admin
		if (isset($_SESSION['sofort']['apiKeyIsValid'])) unset($_SESSION['sofort']['apiKeyIsValid']);
	}
	
	
	function _getGenderFromAddressBook($firstname, $lastname, $company, $streetAddress, $postcode, $city, $countryId, $zoneId) {
		$query = 'SELECT entry_gender
				  FROM	 '.TABLE_ADDRESS_BOOK.'
				  WHERE	 entry_firstname = "'.HelperFunctions::escapeSql($firstname).'"
				  AND	 entry_lastname = "'.HelperFunctions::escapeSql($lastname).'"
				  AND	 entry_company = "'.HelperFunctions::escapeSql($company).'"
				  AND	 entry_street_address = "'.HelperFunctions::escapeSql($streetAddress).'"
				  AND	 entry_postcode = "'.HelperFunctions::escapeSql($postcode).'"
				  AND	 entry_city = "'.HelperFunctions::escapeSql($city).'"
				  AND	 entry_country_id = "'.HelperFunctions::escapeSql($countryId).'"
				  AND	 entry_zone_id = "'.HelperFunctions::escapeSql($zoneId).'" LIMIT 1
		';
		$result = tep_db_fetch_array(tep_db_query($query));
		
		switch ($result['entry_gender']) {
			case 'm': $salutation = 2;  break;
			case 'f': $salutation = 3;  break;
			default : $salutation = ''; break;
		}
		
		return $salutation;
	}
	
	
	/**
	 * check if order has temp-Status and buyer didnt finish the payment-process by "normal" way
	 * @return bool true (always)
	 */
	function _checkCancelOrder() {
		global $cart_pn_ID;
		
		//tmp-order exists, if following if is true
		if (tep_session_is_registered('cart_pn_ID')) {
			$orderId = substr($cart_pn_ID, strpos($cart_pn_ID, '-')+1);
			$pnCartID = substr($cart_pn_ID, 0, strlen($_SESSION['cart']->cartID));
			$checkQuery = tep_db_query("SELECT orders_status FROM ".HelperFunctions::escapeSql(TABLE_ORDERS)." WHERE orders_id = '".(int) $orderId."' LIMIT 1"); //AND payment_method LIKE 'sofort_%'
			$result = tep_db_fetch_array($checkQuery);
			$tempStatus = SOFORT_MULTIPAY_TEMP_STATUS_ID;
			
			if (HelperFunctions::statusIsUnchangedStatus($tempStatus)) {
				if (is_object($this->sofort)) {
					$this->sofort->logWarning("Warning: 'Temporary'-Status is set to 'Unchanged'-status. Don't use this status here and change in modulsettings! Check for 'abort'-status for this order could not be processed. Check order! Order-ID: $orderId");
				} elseif (is_object($this->invoice)) {
					$this->invoice->logWarning("Warning: 'Temporary'-Status is set to 'Unchanged'-status. Don't use this status here and change in modulsettings! Check for 'abort'-status for this order could not be processed. Check order! Order-ID: $orderId");
				}
				return true;
			}
			
			if ($result['orders_status'] == $tempStatus && $_SESSION['cart']->cartID == $pnCartID) {
				tep_session_unregister('cart_pn_ID');
				$this->_cancelOrder( (int) $orderId, 'on');
			}
		}
		
		return true;
	}
	
	
	/**
	 * Check, if needed sofort-lang-constants exists.
	 * If not, include the english-lang-file(s).
	 * @return always true
	 */
	function _checkExistingSofortConstants($paymentMethod) {
		$paymentMethod = strtoupper($paymentMethod);
		$allowedPaymentMethods = array('SU', 'LS', 'SR');
		
		if(!in_array($paymentMethod, $allowedPaymentMethods)) return false;
		
		//security check - constant exists if lang-file exists
		if(defined('SOFORT_'.$paymentMethod.'_TEXT_TITLE')) return true;
		
		$language = HelperFunctions::getSofortLanguage($_SESSION['language']);
		require_once(DIR_FS_CATALOG.DIR_WS_LANGUAGES.$language.'/modules/payment/sofort_'.strtolower($paymentMethod).'.php');
		
		return true;
	}
	
	
	/**
	 * check if order has status "paid" and insert "paid"-status into table_orders and sofort_orders and set comments to history
	 * (sets log-entries in all currious situations)
	 * @return bool TRUE (if errors occured: FALSE)
	 * @todo this function is too long and should be splittet - Notice: this function is the most complicated function of this file!!!
	 */
	function _insertPaidStatus() {
		global $insert_id; //$insert_id equals $orderId!
		
		//create transactionData-Object
		$SofortLib_TransactionData = new SofortLib_TransactionData(SOFORT_MULTIPAY_CONFIGURATION_KEY);
		if (!is_object($SofortLib_TransactionData) || !($SofortLib_TransactionData instanceof SofortLib_TransactionData)) {
			if (is_object($this->invoice)) {
				$this->invoice->logError("TransactionData-Object corrupt. Maybe status could not be correct set! Order-ID: $insert_id. Check status of order!");
			} else {
				$this->sofort->logError("TransactionData-Object corrupt. Maybe status could not be correct set! Order-ID: $insert_id. Check status of order!");
			}
			
			return false;
		}
		
		//get transactionID from table orders
		$transactionId = HelperFunctions::getTransactionIdFromShop($insert_id);
		
		if (!$transactionId){
			$SofortLib_TransactionData->logError("Transaction-ID not found for order-ID: $insert_id. Maybe status could not be correct set! Check status of order!");
			return false;
		}
		
		//fetch all transactionData from SOFORT
		$SofortLib_TransactionData->setTransaction($transactionId);
		$SofortLib_TransactionData->sendRequest();
		
		if ($SofortLib_TransactionData->isError()) {
			$SofortLib_TransactionData->logError("Error in TransactionData: ".print_r($SofortLib_TransactionData->getError(), true)."Transaction-ID: $transactionId. Order-ID: $insert_id. Maybe status could not be correct set! Check status of order!");
			return false;
		}
		
		//check if SOFORT-paymentSecret is equal to shop-paymentSecret
		$paymentSecret = $SofortLib_TransactionData->getUserVariable(3); //paymentSecret was sent to SOFORT in the 4th user variable
		
		if (!HelperFunctions::checkPaymentSecret($paymentSecret, $transactionId)) {
			$SofortLib_TransactionData->logError("Check of Payment-Secret failed for given Transaction-ID: $transactionId! Checked paymentSecret: $paymentSecret. Order-ID: $insert_id. Check order!");
			return false;
		}
		
		//check if SOFORT-orderId is equal to shop-orderId
		$orderId = HelperFunctions::getOrderId($transactionId);
		
		if (!$orderId || $orderId != $insert_id) {
			$SofortLib_TransactionData->logError("Order-ID not found or not equal with shop-order-ID ($insert_id) for Transaction-ID: $transactionId! Check order!");
			return false;
		}
		
		$time = date('d.m.Y, G:i:s', strtotime($SofortLib_TransactionData->getTime()));
		$statusReason = $SofortLib_TransactionData->getStatusReason();
		$status = $SofortLib_TransactionData->getStatus();
		$paymentMethod = $SofortLib_TransactionData->getPaymentMethod();
		$PnagInvoice = '';
		
		if ($paymentMethod == 'sr') {
			$PnagInvoice = new PnagInvoice(SOFORT_MULTIPAY_CONFIGURATION_KEY, $transactionId);
			
			if ($PnagInvoice->isError()) {
				$SofortLib_TransactionData->logError("Errors while creating invoice-object: ".print_r($PnagInvoice->getErrors(), true)." Transaction-ID: $transactionId, Order-ID: $insert_id. Check order!");
				return false;
			}
		}
		
		$dbEntries = array(
				'amount'		=> $SofortLib_TransactionData->getAmount(),
				'status'		=> $status,
				'status_reason' => $statusReason,
		);
		
		if ($paymentMethod == 'sr') $dbEntries['status_id'] = $PnagInvoice->getState();
		
		$statusShort = HelperFunctions::getShortStatus($paymentMethod, $status, $statusReason);
		$newShopStatus = HelperFunctions::getNewShopStatusId($statusShort, $insert_id, $SofortLib_TransactionData);
		$historyComments = HelperFunctions::getHistoryComments($statusShort, $insert_id, $SofortLib_TransactionData, $transactionId, false, $SofortLib_TransactionData);
		
		//check if notification was faster than call of successURL
		if (HelperFunctions::hasEntriesInTableSofortNotifications($transactionId)) {
			$SofortLib_TransactionData->log("Status of first notification was already set by notification (not by success-URL) for order-ID: $insert_id. There is nothing to do here.");
			return true; //everything was done by notification
		}
		
		//save all relevant data in table sofort_notification
		//this DB-entry also prevents race-conditions between call of successUrl and notification
		HelperFunctions::insertSofortOrdersNotification($transactionId, $dbEntries, $SofortLib_TransactionData);
		
		if ($paymentMethod == 'sr') {
			//if buyer has changed addresses at sofort-wizard -> set new addresses also in shopsystem
			if ($statusShort == 'sr_pen_con_inv') HelperFunctions::updateShopAdresses($SofortLib_TransactionData->getInvoiceAddress(), $SofortLib_TransactionData->getShippingAddress(), $insert_id, $SofortLib_TransactionData);
			
			//sync SOFORT-articles/orderTotals with the shop-articles/orderTotals
			require_once(DIR_FS_CATALOG.'callback/sofort/ressources/scripts/sofortOrderSynchronisation.php');
			$SofortOrderSynchronisation = new sofortOrderSynchronisation();
			$SofortOrderSynchronisation->editArticlesShop($PnagInvoice, $insert_id);
		}
		
		//insert all data into orders-table and orders_status_history-table
		if ($historyComments['buyer']) {
			if (!$newShopStatus) {
				$SofortLib_TransactionData->logError("Status for buyer-comment (".$historyComments['buyer'].") unknown. Comment will not be
						inserted. Check order! Order-Id: $insert_id, Transaction-ID: $transactionId!");
				return false;
			}
			
			HelperFunctions::insertHistoryEntry($insert_id, $newShopStatus, $historyComments['buyer'], '', $time);
		}
		
		if ($historyComments['seller']) HelperFunctions::insertHistoryEntry($insert_id, -1, $historyComments['seller'], '', $time);
		
		//if $newShopStatus is set, set it into table orders
		if ($newShopStatus) HelperFunctions::updateTableOrdersStatus($insert_id, $newShopStatus);
		
		return true;
	}
	
	
	/**
	 * Insert all SOFORT-Status into table orders_status and return them
	 * @return array with all status
	 */
	function _insertAndReturnSofortStatus() {
		require_once('sofortInstall.php');
		
		///SOFORT-langs in this module
		$lngdir = DIR_FS_CATALOG.'includes/languages/';
		$iterator = new DirectoryIterator($lngdir);
		
		foreach ($iterator as $file) if ($file->isDir() && file_exists($lngdir.$file->getFilename().'/modules/payment/sofort_general.php')) $sofortLangs[] = strtoupper($file->getFilename());
		
		//current installed langs
		$installedLangs = array();
		$orderQuery = tep_db_query("select languages_id, directory from " . TABLE_LANGUAGES);
		
		while ($result = tep_db_fetch_array($orderQuery)) $installedLangs[$result['languages_id']] = strtoupper($result['directory']);
		
		//get the current highest orders_status_id
		$orderQuery = tep_db_query("SELECT MAX(orders_status_id) AS max_orders_status_id FROM orders_status");
		$maxOrdersStatusIdTemp = tep_db_fetch_array($orderQuery);
		$ordersStatusIdTemp = $maxOrdersStatusIdTemp['max_orders_status_id'] + 1;
		$ordersStatusIdConfirmed = $ordersStatusIdTemp + 1;
		$ordersStatusIdCanceled = $ordersStatusIdConfirmed + 1;
		$ordersStatusIdCheck = $ordersStatusIdCanceled + 1;
		$ordersStatusIdUnconfirmed = $ordersStatusIdCheck + 1;
		$ordersStatusIdInvoiceConfirmed = $ordersStatusIdUnconfirmed + 1;
		$ordersStatusIdAborted = $ordersStatusIdInvoiceConfirmed + 1;
		$ordersStatusIdReturnedDebit = $ordersStatusIdAborted + 1;
		$ordersStatusIdRefunded = $ordersStatusIdReturnedDebit + 1;
		$ordersStatusIdDirectDebit = $ordersStatusIdRefunded + 1;
		$ordersStatusIdTranslateConfirmed = $ordersStatusIdDirectDebit + 1;
		$ordersStatusIdUnchanged = $ordersStatusIdTranslateConfirmed + 1;
		$sofortStatuses = array();
		
		foreach($installedLangs as $installedLang) {
			//insert english for shop-languages, which are not included in this module
			if(in_array($installedLang, $sofortLangs)) {
				$sofortLang = $installedLang;
				$langId = array_search($sofortLang, $installedLangs);
			} else {
				$sofortLang = 'ENGLISH';
				$langId = array_search($installedLang, $installedLangs);
			}
			
			//insert temp-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('temp', $sofortLang);
			$sofortStatuses['temp'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['temp'] === false || $sofortStatuses['temp'] == '') $sofortStatuses['temp'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdTemp, $langId);
			
			//insert confirmed-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('confirmed', $sofortLang);
			$sofortStatuses['confirmed'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['confirmed'] === false || $sofortStatuses['confirmed'] == '') $sofortStatuses['confirmed'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdConfirmed, $langId);
			
			//insert translate_confirmed-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('translate_confirmed', $sofortLang);
			$sofortStatuses['translate_confirmed'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['translate_confirmed'] === false || $sofortStatuses['translate_confirmed'] == '') $sofortStatuses['translate_confirmed'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdTranslateConfirmed, $langId);
			
			//insert canceled-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('canceled', $sofortLang);
			$sofortStatuses['canceled'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['canceled'] === false || $sofortStatuses['canceled'] == '') $sofortStatuses['canceled'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdCanceled, $langId);
			
			//insert aborted-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('aborted', $sofortLang);
			$sofortStatuses['aborted'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['aborted'] === false || $sofortStatuses['aborted'] == '') $sofortStatuses['aborted'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdAborted, $langId);
			
			//insert check-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('check', $sofortLang);
			$sofortStatuses['check'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['check'] === false || $sofortStatuses['check'] == '') $sofortStatuses['check'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdCheck, $langId);
			
			//insert unconfirmed-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('unconfirmed', $sofortLang);
			$sofortStatuses['unconfirmed'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['unconfirmed'] === false || $sofortStatuses['unconfirmed'] == '') $sofortStatuses['unconfirmed'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdUnconfirmed, $langId);
			
			//insert invoice-confirmed-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('invoice_confirmed', $sofortLang);
			$sofortStatuses['invoice_confirmed'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['invoice_confirmed'] === false || $sofortStatuses['invoice_confirmed'] == '') $sofortStatuses['invoice_confirmed'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdInvoiceConfirmed, $langId);
			
			//insert returned-debit-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('returned_debit', $sofortLang);
			$sofortStatuses['returned_debit'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['returned_debit'] === false || $sofortStatuses['returned_debit'] == '') $sofortStatuses['returned_debit'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdReturnedDebit, $langId);
			
			//insert refunded-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('refunded', $sofortLang);
			$sofortStatuses['refunded'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['refunded'] === false || $sofortStatuses['refunded'] == '') $sofortStatuses['refunded'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdRefunded, $langId);
			
			//insert direct-debit-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('direct_debit', $sofortLang);
			$sofortStatuses['direct_debit'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['direct_debit'] === false || $sofortStatuses['direct_debit'] == '') $sofortStatuses['direct_debit'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdDirectDebit, $langId);
			
			//insert unchanged-Status if not exists
			$newOrdersStatusName = $this->_getNewOrdersStatusName('unchanged', $sofortLang);
			$sofortStatuses['unchanged'] = $this->_getStatusIdIfExistInDb($newOrdersStatusName, $langId);
			
			if($sofortStatuses['unchanged'] === false || $sofortStatuses['unchanged'] == '') $sofortStatuses['unchanged'] = $this->_insertStatusInDb($newOrdersStatusName, $ordersStatusIdUnchanged, $langId);
		}
		
		return $sofortStatuses;
	}
	
	
	function _getStatusIdIfExistInDb($newOrdersStatusName, $langId) {
		if(!$newOrdersStatusName) return false;
		
		$checkQuery = tep_db_query('SELECT orders_status_id
									FROM   orders_status
									WHERE  language_id = "'.HelperFunctions::escapeSql($langId).'"
									AND    orders_status_name = "'.HelperFunctions::escapeSql($newOrdersStatusName).'"
									LIMIT  1');
		
		if (tep_db_num_rows($checkQuery) < 1) {
			return false;
		} else {
			//if string exists, we save the orders_status_id
			$neededOrdersStatusId = tep_db_fetch_array($checkQuery);
			
			return $neededOrdersStatusId['orders_status_id'];
		}
	}
	
	
	/**
	 * insert given statusstring into DB - empty strings will not be inserted
	 * @return $orders_status_id from DB OR false
	 */
	function _insertStatusInDb($newOrdersStatusName, $ordersStatusId, $langId) {
		if (!$newOrdersStatusName) return false;
		
		tep_db_query('INSERT INTO orders_status (orders_status_id, language_id, orders_status_name) VALUES ('.$ordersStatusId.', "'.HelperFunctions::escapeSql($langId).'", "'.HelperFunctions::escapeSql($newOrdersStatusName).'")');
		
		return $ordersStatusId;
	}
	
	
	/**
	 * returns the statusname for the given $status and $lang
	 * return max the first 32 chars! (because db-field = varchar(32) )
	 */
	function _getNewOrdersStatusName($status, $lang) {
		switch ($status) {
			case 'temp': 				$newOrdersStatusName = constant('SOFORT_MULTIPAY_STATE_TEMP_'.$lang); break;
			case 'confirmed':			$newOrdersStatusName = constant('SOFORT_MULTIPAY_STATE_CONFIRMED_'.$lang); break;
			case 'translate_confirmed': $newOrdersStatusName = constant('SOFORT_MULTIPAY_STATE_CONFIRMED_2_'.$lang); break;
			case 'invoice_confirmed':	$newOrdersStatusName = constant('SOFORT_MULTIPAY_STATE_INVOICE_CONFIRMED_'.$lang); break;
			case 'unconfirmed':			$newOrdersStatusName = constant('SOFORT_MULTIPAY_STATE_UNCONFIRMED_'.$lang); break;
			case 'canceled':			$newOrdersStatusName = constant('SOFORT_MULTIPAY_STATE_CANCELED_'.$lang); break;
			case 'aborted':				$newOrdersStatusName = constant('SOFORT_MULTIPAY_STATE_ABORTED_'.$lang); break;
			case 'check':				$newOrdersStatusName = constant('SOFORT_MULTIPAY_STATE_CHECK_'.$lang); break;
			case 'returned_debit':		$newOrdersStatusName = constant('SOFORT_MULTIPAY_STATE_RETURNED_DEBIT_'.$lang); break;
			case 'refunded':			$newOrdersStatusName = constant('SOFORT_MULTIPAY_STATE_REFUNDED_'.$lang); break;
			case 'direct_debit':		$newOrdersStatusName = constant('SOFORT_MULTIPAY_STATE_DIRECT_DEBIT_'.$lang); break;
			case 'unchanged':			$newOrdersStatusName = constant('SOFORT_MULTIPAY_STATE_UNCHANGED_'.$lang); break;
		}
		// if string is not cut to 32 chars, status will be reinserted with every installation
		$newOrdersStatusName = substr($newOrdersStatusName, 0, 32);
		
		return $newOrdersStatusName;
	}
	
	
	function _installSofortOrdersTable() {
		$sql = '
			CREATE TABLE IF NOT EXISTS `sofort_orders` (
				`id` int(11) unsigned NOT NULL AUTO_INCREMENT,
				`orders_id` int(11) unsigned NOT NULL,
				`transaction_id` varchar(32) COLLATE utf8_unicode_ci NOT NULL,
				`payment_method` varchar(32) COLLATE utf8_unicode_ci NOT NULL,
				`payment_secret` varchar(32) COLLATE utf8_unicode_ci NOT NULL,
				`date_time` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,
				PRIMARY KEY (`id`)
			)
		';
		tep_db_query($sql);
	}
	
	
	function _removeSofortOrdersTable() {
		$sql = 'DROP TABLE `sofort_orders`';
		tep_db_query($sql);
	}
	
	
	function _installSofortOrdersNotificationTable() {
		$sql = '
			CREATE TABLE IF NOT EXISTS `sofort_orders_notification` (
				`id` int(11) unsigned NOT NULL AUTO_INCREMENT,
				`sofort_orders_id` int(11) unsigned NOT NULL,
				`items` text COLLATE utf8_unicode_ci NOT NULL,
				`amount` float NOT NULL,
				`customer_comment` text COLLATE utf8_unicode_ci NOT NULL,
				`seller_comment` text COLLATE utf8_unicode_ci NOT NULL,
				`status_id` int(11) unsigned NOT NULL,
				`status` varchar(32) COLLATE utf8_unicode_ci NOT NULL,
				`status_reason` varchar(32) COLLATE utf8_unicode_ci NOT NULL,
				`invoice_status` varchar(32) COLLATE utf8_unicode_ci NOT NULL,
				`invoice_objection` varchar(45) COLLATE utf8_unicode_ci NOT NULL,
				`date_time` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,
				PRIMARY KEY (`id`)
			)
		';
		tep_db_query($sql);
	}
	
	
	function _removeSofortOrdersNotificationTable() {
		$sql = 'DROP TABLE `sofort_orders_notification`';
		tep_db_query($sql);
	}
	
	
	function _installSofortProductsTable() {
		$sql = '
			CREATE TABLE IF NOT EXISTS `sofort_products` (
				`id` int(11) unsigned NOT NULL AUTO_INCREMENT,
				`orders_id` int(11) unsigned NOT NULL,
				`orders_products_id` int(11) unsigned NOT NULL,
				`item_id` text COLLATE utf8_unicode_ci NOT NULL,
				PRIMARY KEY (`id`)
			)
		';
		tep_db_query($sql);
	}
	
	
	function _removeSofortProductsTable() {
		$sql = 'DROP TABLE `sofort_products`';
		tep_db_query($sql);
	}
	
	
	/**
	 * replace special chars in $reason e.g. �=>ae, �=>ss incl. foreign non-ascii-signs etc.
	 * @param string $reason like 'K�ufer A. M�ller'
	 * @return string - converted String
	 */
	function _convertReason($reason) {
		$oldLocale = setlocale(LC_ALL, 0);
		setlocale(LC_ALL, "de_DE.utf8");
		$shopEncoding = HelperFunctions::getIniValue('shopEncoding');
		//$encoding = mb_detect_encoding('aaa'.$reason, "ISO-8859-1, UTF-8");
		$convertedReason = substr(iconv($shopEncoding, "ASCII//TRANSLIT", 'aaa'.$reason),3);
		setlocale(LC_ALL, $oldLocale);
		
		return $convertedReason;
	}
	
	
	/**
	 * get the reasons for the given payment method
	 */
	function _getReasons($paymentMethod, $customerId, $order, $orderId) {
		$reason_1 = str_replace('{{order_id}}', $orderId, SOFORT_MULTIPAY_REASON_1);
		$reason_1 = str_replace('{{customer_id}}', $customerId, $reason_1);
		$reason_1 = str_replace('{{transaction_id}}', '-TRANSACTION-', $reason_1);
		$reason_1 = $this->_convertReason($reason_1);
		$reason_1 = tep_output_string(substr($reason_1, 0, 27));
		$reason_2 = str_replace('{{order_id}}', $orderId, SOFORT_MULTIPAY_REASON_2);
		$reason_2 = str_replace('{{customer_id}}', $customerId, $reason_2);
		$reason_2 = str_replace('{{order_date}}', strftime(DATE_FORMAT_SHORT), $reason_2);
		$reason_2 = str_replace('{{customer_name}}', $order->customer['firstname'] . ' ' . $order->customer['lastname'], $reason_2);
		$reason_2 = str_replace('{{customer_company}}', $order->customer['company'], $reason_2);
		$reason_2 = str_replace('{{customer_email}}', $order->customer['email_address'], $reason_2);
		$reason_2 = str_replace('{{transaction_id}}', '-TRANSACTION-', $reason_2);
		$reason_2 = $this->_convertReason($reason_2);
		$reason_2 = tep_output_string(substr($reason_2, 0, 27));
		$reasons = array();
		$reasons[0] = $reason_1;
		$reasons[1] = $reason_2;
		
		return $reasons;
	}
	
	
	/**
	 * @param bool $restock - only allowed is: FALSE or 'on'
	 */
	function _cancelOrder($orderId, $restock) {
		if ($restock == 'on') {
			$order_query = tep_db_query("SELECT orders_products_id, products_id, products_quantity
										 FROM	".TABLE_ORDERS_PRODUCTS."
										 WHERE	orders_id = '".tep_db_input($orderId)."'");
			
			while ($order = tep_db_fetch_array($order_query)) {
				tep_db_query("UPDATE ".TABLE_PRODUCTS."
							  SET	 products_quantity = products_quantity + ".HelperFunctions::escapeSql($order['products_quantity']).", products_ordered = products_ordered - ".HelperFunctions::escapeSql($order['products_quantity'])."
							  WHERE	 products_id = '".HelperFunctions::escapeSql($order['products_id'])."'");
			}
		}
		
		//update status and customer-history
		$time = date("d.m.Y, G:i:s");
		tep_db_query('UPDATE '.TABLE_ORDERS.' SET orders_status = "'.HelperFunctions::escapeSql(SOFORT_MULTIPAY_ABORTED_STATUS_ID).'", last_modified = now() WHERE orders_id = '.HelperFunctions::escapeSql($orderId));
		HelperFunctions::insertHistoryEntry((int) $orderId, SOFORT_MULTIPAY_ABORTED_STATUS_ID, SOFORT_MULTIPAY_ORDER_CANCELED, '', $time);
	}
	
	
	/**
	 * checks, if current payment-method is installed (installed != enabled)
	 */
	function _isInstalled() {
		$installedPaymentMethods = explode(';', MODULE_PAYMENT_INSTALLED);
		
		return (in_array($this->code.'.php', $installedPaymentMethods)) ? true : false;
	}
	
	
	function cleanString($string) {
		return htmlentities(strip_tags($string), ENT_QUOTES, HelperFunctions::getIniValue('shopEncoding'));
	}
}